# Empirical application based on the dataset from Toivanen and Waterson (2005) on Macdonald's and Burger King in the UK.

rm(list=ls())
setwd("")	# Set working directory here
library("BB")
library("foreach")
library("doParallel")

sFileName <- "multi"

lSink <- TRUE # Saving output to *.txt file?
lParallel <- TRUE # Parallel starting values?
lSaveMatrix <- TRUE # Save matrix?

## Setup parallel computing and sinking.
if (lParallel==TRUE){
  dCores=detectCores()
  dCl <- makeCluster(dCores[1]-1)
  registerDoParallel(dCl)  
}

if (lSink==TRUE){
  writeLines(c(""), paste0("01_out_",sFileName,".txt"))
  sink(paste0("01_out_",sFileName,".txt"))
}

## Part 1: Specification of some constants.

# Number of starting values.
dstartvals <- 100

# Small value used for relaxation parameter.
alpharlx <- 0.05 

# Maximum number of iterations
maxiter <- 1000

# Tolerance level for fixed point in CCP space.
dtol <- 1e-6

# Constants of the model.
maxstore <- 15	# Maximum number of stores
nplayer <- 2	# Number of players
dfact <- 0.95	# Discount factor

# Constants of the datafile.
nyear <- 5	# Number of years
nmarket <- 422	# Number of local markets

# Names of parameters.
namesb1 <- c("VP0_BK", "VP1_BK", "VP2_BK", "FC0_BK", "FC1_BK", "FC2_BK", "VP0_MD", "VP1_MD", "VP2_MD", "FC0_MD", "FC1_MD", "FC2_MD") 
namesb2 <- c("DENSITY", "GDP", "RENT", "T")
namesb <- c(namesb1, namesb2)

# Calculating some constants
vstate <- seq(0, maxstore-1)
vstate <- cbind(vstate%x%rep(1,maxstore), rep(1,maxstore)%x%vstate)
nstate <- nrow(vstate)


## Part 2: Reading data and construction of variables

df <- readRDS("bkmd.rda")

# Construction of variables.
x_bk <- df$bk_stock	# Stock of stores for BK
x_md <- df$mcd_stock	# Stock of stores for MD
a_bk <-  df$bk_entry>0	# Dummy of new entry for BK
a_md <- df$mcd_entry>0	# Dummy of new entry for MD
df$population <- df$population/1000	# Population in millions
df$density <- df$population/df$district_area

# Market specific mean values of some exogenous explanatory variables.
dfMkt <- aggregate(df[,c("population", "density", "gdp_pc", "avg_rent", "ctax")], list("district_code"=df$district_code), mean)
marketsize <- dfMkt$population
zmarket <- as.matrix(dfMkt[,c("density", "gdp_pc", "avg_rent", "ctax")])


## Part 3: Procedures.

freqprob <- function(YOBS, XOBS, XVAL){
	NUMX <- nrow(XVAL)
	NUMQ <- ncol(YOBS)
	PROB1 <- matrix(0, NUMX, NUMQ)
	t <- 1
	
	while (t<=NUMX){
		SELX <- apply(XOBS==matrix(XVAL[t,], nrow(XOBS), ncol(XVAL), byrow=TRUE), 1, prod)
		DENOM <- sum(SELX)
		if (DENOM==0){
			PROB1[t,] <- rep(0, NUMQ)
		} else{
			NUMER <- colSums(SELX*YOBS)
			PROB1[t,] <- t(NUMER)/DENOM
		}
		t <- t+1
	}

	return(PROB1)
}

miprobit <- function(YDUM, X, REST, B0, NAMES, PRINT, VAR){
	MYZERO <- 1e-16
	NOBS <- length(YDUM)
	NPARAM <- ncol(X)
	EPS <- 1e-6
	ITER <- 1
	LLIKE <- 1000
	CRITER <- 1000

	while (CRITER>EPS){
		if (PRINT){
			cat("\n")
			cat("Pseudo MLE Iteration = ", ITER, "\n")
			cat("Log-Likelihood funciton = ", LLIKE, "\n")
			cat("Criterion = ", CRITER, "\n")
			cat("\n")
		}
		FXB0 <- pnorm(X%*%B0 + REST)
		FXB0 <- FXB0 + (MYZERO - FXB0)*(FXB0<MYZERO) + (1 - MYZERO - FXB0)*(FXB0>(1 - MYZERO))
		LLIKE <- t(YDUM)%*%log(FXB0) + t(1-YDUM)%*%log(1-FXB0)
		PHIXB0 <- dnorm(X%*%B0 + REST)
		LAMBDAB0 <- YDUM*(PHIXB0/FXB0) + (1-YDUM)*(-PHIXB0/(1-FXB0))
		DLOGLB0 <- t(X)%*%LAMBDAB0
		D2LOGLB0 <- -t(as.vector(LAMBDAB0*(LAMBDAB0 + X%*%B0 + REST))*X)%*%X
		B1 <- B0 - solve(D2LOGLB0)%*%DLOGLB0
		CRITER <- max(abs(B1-B0))
		B0 <- B1
		ITER <- ITER + 1
	}

	if (VAR){
		FXB0 <- pnorm(X%*%B0 + REST)
		FXB0 <- FXB0 + (MYZERO - FXB0)*(FXB0<MYZERO) + (1 - MYZERO - FXB0)*(FXB0>(1 - MYZERO))
		LLIKE <- t(YDUM)%*%log(FXB0) + t(1-YDUM)%*%log(1-FXB0)
		PHIXB0 <- dnorm(X%*%B0 + REST)
		LAMBDA0 <- -PHIXB0/(1-FXB0)
		LAMBDA1 <- PHIXB0/FXB0
		AVARB <- t(as.vector(LAMBDA0*LAMBDA1)*X)%*%X
		AVARB <- solve(-AVARB)
		SDB <- sqrt(diag(AVARB))
		TSTAT <- B0/SDB
		NUMY1 <- sum(YDUM)
		NUMY0 <- NOBS - NUMY1
		LOGL0 <- NUMY1*log(NUMY1) + NUMY0*log(NUMY0) - NOBS*log(NOBS)
		LRI <- 1- LLIKE/LOGL0
		PSEUDOR2 <- 1 - (t(YDUM - FXB0)%*%(YDUM - FXB0))/NUMY1		
	} else{
		AVARB <- NA
	}

	if (PRINT){
		MPRINT <- round(cbind(B0, SDB, TSTAT), 5)
		colnames(MPRINT) <- c("Estimate", "Std Errors", "t-ratios")
		rownames(MPRINT) <- NAMES
		cat("Number of iterations = ", ITER, "\n")
		cat("Log-Likelihood function = ", LLIKE, "\n")
		cat("Likelihood Ratio Index = ", LRI, "\n")
		cat("Pseudo-R2 = ", PSEUDOR2, "\n")
		print(MPRINT)
	}

	OUTMIPROBIT <- list("thetaest" = B0, "varest" = AVARB, "likelihood" = LLIKE)
	return(OUTMIPROBIT)
}

nplmap <- function(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, NAMESB, PRINT, VAR){
	# Some constants.
	MYZERO <- 1e-12
	NOBS <- nrow(YOBS)
	NMARKET <- length(MSIZE)
	NYEAR <- NOBS/NMARKET
	if (NYEAR%%1!=0) cat("\n ERROR: Number of years is not an integer \n")
	NPLAYER <- ncol(YOBS)
	NS <- nrow(PCHOICE)/NMARKET
	if (NS%%1!=0) cat("\n ERROR: Number of states in a single market is not an integer\n")
	NUMX <- sqrt(NS)
	if (NUMX%%1!=0) cat("\n ERROR: Number of values of xbk or xmd is not an integer\n")
	KTOT <- length(NAMESB)
	KZ <- ncol(ZMARKET)
	KVPFC <- (KTOT-KZ)/2
	if (KVPFC%%1!=0) cat("\n ERROR: Number of parameters in var profits and fixed costs is not an integer\n")
	XBK <- MSTATE[,1]
	XMD <- MSTATE[,2]

	# Vector with indexes for the observed state.
	INDXOBS <- rep(0, NOBS)
	j <- 1
	while (j <= NS){
		INDXOBS <- INDXOBS + j*apply(XOBS==matrix(MSTATE[j,], nrow(XOBS), ncol(MSTATE), byrow=TRUE), 1, prod)
		j <- j+1
	}

	ZTILDE_BK <- matrix(0, NMARKET*NS, KVPFC+KZ)
	ZTILDE_MD <- matrix(0, NMARKET*NS, KVPFC+KZ)
	ETILDE_BK <- matrix(0, NMARKET*NS, 1)
	ETILDE_MD <- matrix(0, NMARKET*NS, 1)
	ZTILDE_OBS_BK <- matrix(0, NOBS, KVPFC+KZ)
	ZTILDE_OBS_MD <- matrix(0, NOBS, KVPFC+KZ)
	ETILDE_OBS_BK <- matrix(0, NOBS, 1)
	ETILDE_OBS_MD <- matrix(0, NOBS, 1)	

	m <- 1
	while (m <= NMARKET){
		VALMSIZE <- MSIZE[m]
		VALZMARKET <- ZMARKET[m,]

		# Truncated probabilities to avoid numerical issues.
		P_BK <- PCHOICE[((m-1)*NS+1):(m*NS),1]
		P_MD <- PCHOICE[((m-1)*NS+1):(m*NS),2]
		P_BK <- (P_BK <= MYZERO)*MYZERO + (P_BK >= 1-MYZERO)*(1-MYZERO) + (P_BK >= MYZERO)*(P_BK <= 1-MYZERO)*P_BK
		P_MD <- (P_MD <= MYZERO)*MYZERO + (P_MD >= 1-MYZERO)*(1-MYZERO) + (P_MD >= MYZERO)*(P_MD <= 1-MYZERO)*P_MD

		# Vectors of expected profits.
		ZBK_00 <- cbind(VALMSIZE*(XBK>0), VALMSIZE*(XBK-XMD), VALMSIZE*(XBK-XMD)^2, -(XBK>0), -XBK, -XBK^2, XBK%*%t(VALZMARKET))
		ZBK_01 <- cbind(VALMSIZE*(XBK>0), VALMSIZE*(XBK-XMD-1), VALMSIZE*(XBK-XMD-1)^2, -(XBK>0), -XBK, -XBK^2, XBK%*%t(VALZMARKET))
		ZBK_10 <- cbind(VALMSIZE*(XBK+1>0), VALMSIZE*(XBK+1-XMD), VALMSIZE*(XBK+1-XMD)^2, -(XBK+1>0), -(XBK+1), -(XBK+1)^2, (XBK+1)%*%t(VALZMARKET))
		ZBK_11 <- cbind(VALMSIZE*(XBK+1>0), VALMSIZE*(XBK+1-XMD-1), VALMSIZE*(XBK+1-XMD-1)^2, -(XBK+1>0), -(XBK+1), -(XBK+1)^2, (XBK+1)%*%t(VALZMARKET))

		ZMD_00 <- cbind(VALMSIZE*(XMD>0), VALMSIZE*(XMD-XBK), VALMSIZE*(XMD-XBK)^2, -(XMD>0), -XMD, -XMD^2, XMD%*%t(VALZMARKET))
		ZMD_01 <- cbind(VALMSIZE*(XMD>0), VALMSIZE*(XMD-XBK-1), VALMSIZE*(XMD-XBK-1)^2, -(XMD>0), -XMD, -XMD^2, XMD%*%t(VALZMARKET))
		ZMD_10 <- cbind(VALMSIZE*(XMD+1>0), VALMSIZE*(XMD+1-XBK), VALMSIZE*(XMD+1-XBK)^2, -(XMD+1>0), -(XMD+1), -(XMD+1)^2, (XMD+1)%*%t(VALZMARKET))
		ZMD_11 <- cbind(VALMSIZE*(XMD+1>0), VALMSIZE*(XMD+1-XBK-1), VALMSIZE*(XMD+1-XBK-1)^2, -(XMD+1>0), -(XMD+1), -(XMD+1)^2, (XMD+1)%*%t(VALZMARKET))

		EPROFBK_0 <- (1-P_MD)*ZBK_00 + P_MD*ZBK_01 	# Expected profit of BK if a=0
		EPROFBK_1 <- (1-P_MD)*ZBK_10 + P_MD*ZBK_11 	# Expected profit of BK if a=1
		EPROFMD_0 <- (1-P_BK)*ZMD_00 + P_BK*ZMD_01 	# Expected profit of MD if a=0
		EPROFMD_1 <- (1-P_BK)*ZMD_10 + P_BK*ZMD_11 	# Expected profit of MD if a=1

		# Transition probabilities.

		# Transition xbk for BK with abk=0: Pr( xbk' | xbk, xmd, abk =0) = 1{xbk' = xbk)
		TRANXBK_BK0 <- diag(NUMX)	# Transition of xbk in the space of xbk
		TRANXBK_BK0 <- TRANXBK_BK0%x%matrix(1, NUMX, NUMX)	# Transition of xbk in the space of xbk, xmd

		# Transition xbk for BK with abk=1: Pr( xbk' | xbk, xmd, abk =1) = 1{xbk' = xbk+1)
		TRANXBK_BK1 <- rbind(cbind(rep(0,NUMX-1), diag(NUMX-1)), t(c(rep(0,NUMX-1),1)))	# Transition of xbk in the space of xbk
		TRANXBK_BK1 <- TRANXBK_BK1%x%matrix(1, NUMX, NUMX)	# Transition of xbk in the space of xbk, xmd

		# Transition xmd for MD with amd=0: Pr( xmd' | xbk, xmd, amd =0) = 1{xmd' = xmd)
		TRANXMD_MD0 <- diag(NUMX)	# Transition of xmd in the space of xmd
		TRANXMD_MD0 <- matrix(1, NUMX, NUMX)%x%TRANXMD_MD0	# Transition of xmd in the space of xbk, xmd

		# Transition xmd for MD with amd=1: Pr( xmd' | xbk, xmd, amd =1) = 1{xmd' = xmd+1)
		TRANXMD_MD1 <- rbind(cbind(rep(0,NUMX-1), diag(NUMX-1)), t(c(rep(0,NUMX-1),1)))	# Transition of xmd in the space of xmd
		TRANXMD_MD1 <- matrix(1, NUMX, NUMX)%x%TRANXMD_MD1	# Transition of xmd in the space of xbk, xmd

		# Transition xmd from the point of view of BK who does not know amd: Pr( xmd' | xbk, xmd) = (1-pmd) * 1{xmd' = xmd) +  pmd * 1{xmd' = xmd+1)
		TRANXMD_BK <- (1-P_MD)*TRANXMD_MD0 + P_MD*TRANXMD_MD1

		# Transition xbk from the point of view of MD who does not know abk: Pr( xbk' | xbk, xmd) = (1-pbk) * 1{xbk' = xbk) +  pbk * 1{xbk' = xbk+1)
		TRANXBK_MD <- (1-P_BK)*TRANXBK_BK0 + P_BK*TRANXBK_BK1

		# Checking if transition matrices sum to 1.
		
		# Total transition matrix of (xbk,xmd) for BK if abk = 0
		# Pr( xbk',xmd' | xbk, xmd, abk=0) = Pr( xbk' | xbk, abk =0) *  Pr( xmd' | xbk, xmd)
	    TOTTRAN_BK0 <- TRANXBK_BK0*TRANXMD_BK
	    if (any(rowSums(TOTTRAN_BK0) > 1.00001) | any(rowSums(TOTTRAN_BK0) < 0.99999)) cat("\n ERROR: Transition matrix does not sum to 1\n")

		# Total transition matrix of (xbk,xmd) for BK if abk = 1
  		# Pr( xbk',xmd' | xbk, xmd, abk=1) = Pr( xbk' | xbk, abk =1) *  Pr( xmd' | xbk, xmd)
	    TOTTRAN_BK1 <- TRANXBK_BK1*TRANXMD_BK
	    if (any(rowSums(TOTTRAN_BK1) > 1.00001) | any(rowSums(TOTTRAN_BK1) < 0.99999)) cat("\n ERROR: Transition matrix does not sum to 1\n")

	   	# Total transition matrix of (xbk,xmd) for MD if amd = 0
  		# Pr( xbk',xmd' | xbk, xmd, amd=0) = Pr( xmd' | xmd, amd =0) *  Pr( xbk' | xbk, xmd)
	    TOTTRAN_MD0 <- TRANXMD_MD0*TRANXBK_MD
	    if (any(rowSums(TOTTRAN_MD0) > 1.00001) | any(rowSums(TOTTRAN_MD0) < 0.99999)) cat("\n ERROR: Transition matrix does not sum to 1\n")

	   	# Total transition matrix of (xbk,xmd) for MD if amd = 1
	   	# Pr( xbk',xmd' | xbk, xmd, amd=1) = Pr( xmd' | xmd, amd =1) *  Pr( xbk' | xbk, xmd)
	    TOTTRAN_MD1 <- TRANXMD_MD1*TRANXBK_MD
	    if (any(rowSums(TOTTRAN_MD1) > 1.00001) | any(rowSums(TOTTRAN_MD1) < 0.99999)) cat("\n ERROR: Transition matrix does not sum to 1\n")

	    # Unconditional transition matrix
	    UNCONTRAN <- (1-P_BK)*TOTTRAN_BK0 + P_BK*TOTTRAN_BK1
		if (any(rowSums(UNCONTRAN) > 1.00001) | any(rowSums(UNCONTRAN) < 0.99999)) cat("\n ERROR: Transition matrix does not sum to 1\n")

		# ztilde and etilde for every possible states
		UNCONTRAN <- solve(diag(NS) - BETA*UNCONTRAN)	# Matrix (I - beta*F)^-1
		VALUE_Z_BK <- (1-P_BK)*EPROFBK_0 + P_BK*EPROFBK_1
		VALUE_Z_BK <- UNCONTRAN%*%VALUE_Z_BK	# Value Z function BK
		VALUE_E_BK <- UNCONTRAN%*%dnorm(qnorm(P_BK))	# Value e function BK
		VALUE_Z_MD <- (1-P_MD)*EPROFMD_0 + P_MD*EPROFMD_1
		VALUE_Z_MD <- UNCONTRAN%*%VALUE_Z_MD	# Value Z function MD
		VALUE_E_MD <- UNCONTRAN%*%dnorm(qnorm(P_MD))	# Value e function MD

		ZT_BK <- (EPROFBK_1 - EPROFBK_0) + BETA*(TOTTRAN_BK1 - TOTTRAN_BK0)%*%VALUE_Z_BK
		ZT_MD <- (EPROFMD_1 - EPROFMD_0) + BETA*(TOTTRAN_MD1 - TOTTRAN_MD0)%*%VALUE_Z_MD
		ET_BK <- BETA*(TOTTRAN_BK1 - TOTTRAN_BK0)%*%VALUE_E_BK
		ET_MD <- BETA*(TOTTRAN_MD1 - TOTTRAN_MD0)%*%VALUE_E_MD

		COUNT1 <- (m-1)*NS + 1
		COUNT2 <- m*NS
		ZTILDE_BK[COUNT1:COUNT2,] <- ZT_BK
		ZTILDE_MD[COUNT1:COUNT2,] <- ZT_MD
		ETILDE_BK[COUNT1:COUNT2,] <- ET_BK
		ETILDE_MD[COUNT1:COUNT2,] <- ET_MD

		COUNT1 <- (m-1)*NYEAR + 1
		COUNT2 <- m*NYEAR
		ZTILDE_OBS_BK[COUNT1:COUNT2,] <- ZT_BK[INDXOBS[COUNT1:COUNT2],]
		ZTILDE_OBS_MD[COUNT1:COUNT2,] <- ZT_MD[INDXOBS[COUNT1:COUNT2],]
		ETILDE_OBS_BK[COUNT1:COUNT2,] <- ET_BK[INDXOBS[COUNT1:COUNT2],]
		ETILDE_OBS_MD[COUNT1:COUNT2,] <- ET_MD[INDXOBS[COUNT1:COUNT2],]

		m <- m+1
	}

	# Pseudo Maximum Likelihood Estimation
	ZOBS <- cbind(rbind(ZTILDE_OBS_BK[, 1:KVPFC], matrix(0, NOBS, KVPFC)), rbind(matrix(0, NOBS, KVPFC), ZTILDE_OBS_MD[, 1:KVPFC]), rbind(ZTILDE_OBS_BK[, (KVPFC+1):(KVPFC+KZ)], ZTILDE_OBS_MD[, (KVPFC+1):(KVPFC+KZ)]))	
	EOBS <- rbind(ETILDE_OBS_BK, ETILDE_OBS_MD)
	OUTMIPROBIT <- miprobit(c(YOBS[,1], YOBS[,2]), ZOBS, EOBS, matrix(0, KTOT, 1), NAMESB, PRINT, VAR)

	# Updating conditional choice probabilities.
	THETA_BK <- c(OUTMIPROBIT$thetaest[1:KVPFC], OUTMIPROBIT$thetaest[(2*KVPFC+1):KTOT])
	THETA_MD <- c(OUTMIPROBIT$thetaest[(KVPFC+1):(2*KVPFC)], OUTMIPROBIT$thetaest[(2*KVPFC+1):KTOT])
	NPLPCHOICE <- pnorm(cbind(ZTILDE_BK%*%THETA_BK + ETILDE_BK, ZTILDE_MD%*%THETA_MD + ETILDE_MD))	
	NPLPEST_OBS <- rbind(pnorm(ZTILDE_OBS_BK%*%THETA_BK + ETILDE_OBS_BK), pnorm(ZTILDE_OBS_MD%*%THETA_MD + ETILDE_OBS_MD))

	OUTNPLMAP <- list("nplpchoice"=NPLPCHOICE, "npltheta"=OUTMIPROBIT$thetaest, "npllikelihood"=OUTMIPROBIT$likelihood, "nplvarest"=OUTMIPROBIT$varest, "nplpest_obs"=NPLPEST_OBS)
	return(OUTNPLMAP)
}

npl_bkmd <- function(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, KITER, NAMESB, PRINTIT){
	CRITERION <- 1000
	CONV_CONST <- dtol
	P0 <- PCHOICE
	THETA0 <- rep(0, length(NAMESB))
	NPLITER <- 1 

	while (NPLITER <= KITER & CRITERION > CONV_CONST){
		if (PRINTIT) cat("\n NPL ITERATION = ", NPLITER, ", Criterion = ", CRITERION, "\n")

		# Compute the NPL mapping.
		OUTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, P0, MSTATE, BETA, NAMESB, FALSE, FALSE)
		PCHOICE <- OUTNPLMAP$nplpchoice
		
		# Checking for convergence.
		CRITERION <- max(abs(P0 - PCHOICE))

		P0 <- PCHOICE
		THETA0 <- OUTNPLMAP$npltheta
		NPLITER <- NPLITER+1
	}

	if (CRITERION <= CONV_CONST){
		OUTLASTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, NAMESB, FALSE, TRUE)	
	} else{
		OUTLASTNPLMAP <- list("nplpchoice"=matrix(NA, nrow(PCHOICE), ncol(PCHOICE)), "npltheta"=rep(NA,length(NAMESB)), "npllikelihood"=NA, "nplvarest"=matrix(NA, length(NAMESB), length(NAMESB)), "nplpest_obs"=rep(NA, length(YOBS)))
	}
	
	OUTNPL_BKMD <- list("thetaest"=OUTLASTNPLMAP$npltheta, "varest"=OUTLASTNPLMAP$nplvarest, "pchoice"=OUTLASTNPLMAP$nplpchoice, "pest_obs"=OUTLASTNPLMAP$nplpest_obs, "loglikelihood"=OUTLASTNPLMAP$npllikelihood)
}

rlx_bkmd <- function(ALPHA, YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, KITER, NAMESB, PRINTIT){
	CRITERION <- 1000
	CONV_CONST <- dtol
	P0 <- PCHOICE
	THETA0 <- rep(0, length(NAMESB))
	RLXITER <- 1 

	while (RLXITER <= KITER & CRITERION > CONV_CONST){
		if (PRINTIT) cat("\n RELAXATION ITERATION = ", RLXITER, ", Criterion = ", CRITERION, "\n")

		# Compute the NPL mapping.
		OUTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, P0, MSTATE, BETA, NAMESB, FALSE, FALSE)
		PCHOICE <- (OUTNPLMAP$nplpchoice^ALPHA)*(P0^(1-ALPHA))
		
		# Checking for convergence.
		CRITERION <- max(abs(P0 - PCHOICE))

		P0 <- PCHOICE
		THETA0 <- OUTNPLMAP$npltheta
		RLXITER <- RLXITER+1
	}

	if (CRITERION <= CONV_CONST){
		OUTLASTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, NAMESB, FALSE, TRUE)	
	} else{
		OUTLASTNPLMAP <- list("nplpchoice"=matrix(NA, nrow(PCHOICE), ncol(PCHOICE)), "npltheta"=rep(NA,length(NAMESB)), "npllikelihood"=NA, "nplvarest"=matrix(NA, length(NAMESB), length(NAMESB)), "nplpest_obs"=rep(NA, length(YOBS)))
	}
	
	OUTRLX_BKMD <- list("thetaest"=OUTLASTNPLMAP$npltheta, "varest"=OUTLASTNPLMAP$nplvarest, "pchoice"=OUTLASTNPLMAP$nplpchoice, "pest_obs"=OUTLASTNPLMAP$nplpest_obs, "loglikelihood"=OUTLASTNPLMAP$npllikelihood)
}

phimap <- function(VECPCHOICE, YOBS, XOBS, MSIZE, ZMARKET, MSTATE, BETA, NAMESB){
	PCHOICE <- matrix(VECPCHOICE, length(VECPCHOICE)/ncol(YOBS), ncol(YOBS))
	OUTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, NAMESB, FALSE, FALSE)
	return(c(PCHOICE - OUTNPLMAP$nplpchoice))
}

spec_bkmd <- function(YOBS, XOBS, MSIZE, ZMARKET, PCHOICE, MSTATE, BETA, NAMESB){
	CONV_CONST <- dtol
	OUTSPECSOLVE <- BBsolve(c(PCHOICE), phimap, method=c(2,3,1), control=list(maxit=maxiter, tol = CONV_CONST), quiet=TRUE, YOBS=YOBS, XOBS=XOBS, MSIZE=MSIZE, ZMARKET=ZMARKET, MSTATE=MSTATE, BETA=BETA, NAMESB=NAMESB)
	if (OUTSPECSOLVE$convergence==0){
		SPECPCHOICE <- matrix(OUTSPECSOLVE$par, length(OUTSPECSOLVE$par)/ncol(YOBS), ncol(YOBS))
		OUTLASTNPLMAP <- nplmap(YOBS, XOBS, MSIZE, ZMARKET, SPECPCHOICE, MSTATE, BETA, NAMESB, FALSE, TRUE)
	} else{
		OUTLASTNPLMAP <- list("nplpchoice"=matrix(NA, nrow(PCHOICE), ncol(PCHOICE)), "npltheta"=rep(NA,length(NAMESB)), "npllikelihood"=NA, "nplvarest"=matrix(NA, length(NAMESB), length(NAMESB)), "nplpest_obs"=rep(NA, length(YOBS)))
	}
	OUTSPEC_BKMD <- list("thetaest"=OUTLASTNPLMAP$npltheta, "varest"=OUTLASTNPLMAP$nplvarest, "pchoice"=OUTLASTNPLMAP$nplpchoice, "pest_obs"=OUTLASTNPLMAP$nplpest_obs, "loglikelihood"=OUTLASTNPLMAP$npllikelihood)
}

## Part 4: Estimation of initial CCPs

prob_freq <- matrix(0, nmarket*nstate, nplayer)
market <- 1
while (market <= nmarket){
	count1 <- (market-1)*nstate + 1
	count2 <- market*nstate
	yyy <- cbind(a_bk[((market-1)*nyear+1):(market*nyear)], a_md[((market-1)*nyear+1):(market*nyear)])
	xxx <- cbind(x_bk[((market-1)*nyear+1):(market*nyear)], x_md[((market-1)*nyear+1):(market*nyear)])
	prob_freq[count1:count2,] <- freqprob(yyy, xxx, vstate)
	market <- market+1
}


cat("\n");
cat("Progress of the multiple starting values: \n")
if (lSink==TRUE) sink()

sub <- 10
submulti <- seq(0,dstartvals,dstartvals/sub)

for (ss in 1:(length(submulti)-1)){
	mSubMultiMatrix <-foreach(s=(submulti[ss]+1):submulti[ss+1], .combine=cbind, .packages=c('BB')) %dopar% {  
		if (lSink==TRUE) sink(paste0("01_out_",sFileName,".txt"), append=TRUE)
		cat(paste("\t ", (s/dstartvals)*100, "% \n"))
		set.seed(1234+s)

		if (s==1){
			p0 <- prob_freq
		} else{
			p0 <- matrix(runif(length(prob_freq)), nrow(prob_freq), ncol(prob_freq))
		}

		dstartnpl <- Sys.time()
	    outnpl_bkmd <- npl_bkmd(cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, maxiter, namesb, FALSE)
	    dendnpl <- Sys.time()
	    if (!is.na(outnpl_bkmd$loglikelihood)){
	    	dtimenpl <- difftime(dendnpl, dstartnpl, units="secs")	
	    } else{
	    	dtimenpl <- NA
	    }
	    storenpl <- c(outnpl_bkmd$loglikelihood, outnpl_bkmd$thetaest, dtimenpl)

		dstartrlx <- Sys.time()
	    outrlx_bkmd <- rlx_bkmd(alpharlx, cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, maxiter, namesb, FALSE)
	    dendrlx <- Sys.time()
	    if (!is.na(outrlx_bkmd$loglikelihood)){
	    	dtimerlx <- difftime(dendrlx, dstartrlx, units="secs")	
	    } else{
	    	dtimerlx <- NA
	    }
	    storerlx <- c(outrlx_bkmd$loglikelihood, outrlx_bkmd$thetaest, dtimerlx)

		dstartspec <- Sys.time()
	    outspec_bkmd <- spec_bkmd(cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, namesb)
	    dendspec <- Sys.time()
		if (!is.na(outspec_bkmd$loglikelihood)){
	    	dtimespec <- difftime(dendspec, dstartspec, units="secs")	
	    } else{
	    	dtimespec <- NA
	    }
	    storespec <- c(outspec_bkmd$loglikelihood, outspec_bkmd$thetaest, dtimespec)

	  	if (lSink==TRUE) sink()
	    
	    invisible(c(storenpl, storerlx, storespec))
	}

	if (lSink==TRUE) sink(paste0("01_out_",sFileName,".txt"), append=TRUE)

	mSubMultiMatrix <- as.matrix(mSubMultiMatrix)
	if (ss==1){
	    mMultiMatrix <- mSubMultiMatrix
	} else {
	    mMultiMatrix <- cbind(mMultiMatrix, mSubMultiMatrix)  
	} 

  	if (lSaveMatrix==TRUE) save(mMultiMatrix, file=paste0("02_mat_",sFileName,"_",ss*sub,"pct.rda")) 
  	cat("\n Summary of the loglikelihood for each method after", ss*sub, "% of the ", dstartvals, "starting values \n")
  	mLLikes <- cbind(mMultiMatrix[1,], mMultiMatrix[(nrow(mMultiMatrix)/3+1),], mMultiMatrix[(nrow(mMultiMatrix)*2/3+1),])
  	colnames(mLLikes) <- c("NPL", "Relaxation", "Spectral")
  	print(summary(mLLikes, na.rm=TRUE))

  	cat("\n Summary of the computational time for each method after", ss*sub, "% of the ", dstartvals, "starting values \n")
  	mCompTime <- cbind(mMultiMatrix[nrow(mMultiMatrix)/3,], mMultiMatrix[nrow(mMultiMatrix)*2/3,], mMultiMatrix[(nrow(mMultiMatrix)),])
  	colnames(mCompTime) <- c("NPL", "Relaxation", "Spectral")
  	print(summary(mCompTime, na.rm=TRUE))  	

  	cat("\n Occurrences of failures to converge for each method after", ss*sub, "% of the ", dstartvals, "starting values \n")
  	mFails <- cbind(sum(is.na(mMultiMatrix[1,])), sum(is.na(mMultiMatrix[(nrow(mMultiMatrix)/3+1),])), sum(is.na(mMultiMatrix[(nrow(mMultiMatrix)*2/3+1),])))
  	colnames(mFails) <- c("NPL", "Relaxation", "Spectral")
  	rownames(mFails) <- "Total"
  	print(mFails)
}

cat("\n Best estimates from NPL algorithm after", dstartvals, "starting values: \n")
dbestnpl <- which.max(mLLikes[,1])
set.seed(1234+dbestnpl)
if (dbestnpl==1){
	p0 <- prob_freq
} else{
	p0 <- matrix(runif(length(prob_freq)), nrow(prob_freq), ncol(prob_freq))	
}
outbestnpl_bkmd <- npl_bkmd(cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, maxiter, namesb, FALSE)
bestthetaNPL <- outbestnpl_bkmd$thetaest
beststdNPL <- sqrt(diag(outbestnpl_bkmd$varest))
besttstatNPL <- bestthetaNPL/beststdNPL
bestllikeNPL <- outbestnpl_bkmd$loglikelihood
cat("\n Log-likelihood: ", bestllikeNPL, "\n")
mPrintNPL <- round(cbind(bestthetaNPL, beststdNPL, besttstatNPL), 5)
colnames(mPrintNPL) <- c("Estimate", "Std Errors", "t-ratios")
rownames(mPrintNPL) <- namesb
print(mPrintNPL)

cat("\n")

cat("\n Best estimates from relaxation algorithm after", dstartvals, "starting values: \n")
dbestrlx <- which.max(mLLikes[,2])
set.seed(1234+dbestrlx)
if (dbestrlx==1){
	p0 <- prob_freq
} else{
	p0 <- matrix(runif(length(prob_freq)), nrow(prob_freq), ncol(prob_freq))	
}
outbestrlx_bkmd <- rlx_bkmd(alpharlx, cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, maxiter, namesb, FALSE)
bestthetaRLX <- outbestrlx_bkmd$thetaest
beststdRLX <- sqrt(diag(outbestrlx_bkmd$varest))
besttstatRLX <- bestthetaRLX/beststdRLX
bestllikeRLX <- outbestrlx_bkmd$loglikelihood
cat("\n Log-likelihood: ", bestllikeRLX, "\n")
mPrintRLX <- round(cbind(bestthetaRLX, beststdRLX, besttstatRLX), 5)
colnames(mPrintRLX) <- c("Estimate", "Std Errors", "t-ratios")
rownames(mPrintRLX) <- namesb
print(mPrintRLX)

cat("\n")

cat("\n Best estimates from spectral solver after", dstartvals, "starting values: \n")
dbestspec <- which.max(mLLikes[,3])
set.seed(1234+dbestspec)
if (dbestspec==1){
	p0 <- prob_freq
} else{
	p0 <- matrix(runif(length(prob_freq)), nrow(prob_freq), ncol(prob_freq))	
}
outbestspec_bkmd <- spec_bkmd(cbind(a_bk,a_md), cbind(x_bk,x_md), marketsize, zmarket, p0, vstate, dfact, namesb)
bestthetaSpec <- outbestspec_bkmd$thetaest
beststdSpec <- sqrt(diag(outbestspec_bkmd$varest))
besttstatSpec <- bestthetaSpec/beststdSpec
bestllikeSpec <- outbestspec_bkmd$loglikelihood
cat("\n Log-likelihood: ", bestllikeSpec, "\n")
mPrintSpec <- round(cbind(bestthetaSpec, beststdSpec, besttstatSpec), 5)
colnames(mPrintSpec) <- c("Estimate", "Std Errors", "t-ratios")
rownames(mPrintSpec) <- namesb
print(mPrintSpec)


if (lSink==TRUE) sink()
if (lParallel==TRUE) stopCluster(dCl)

system("say Fini")  