# Simulation of a dynamic game of market entry and exit using data generating processes from Pesendorfer and Schmidt-Dengler (2008).
# Estimation: NPL, spectral algorithms and modified NPL from Kasahara and Shimotsu (2012).
# First step: kernel or frequency count (can set true CCPs to check for coding errors).
# Criterion function: maximum likelihood.

rm(list=ls())
setwd("") # Set working directory here
library("MASS")
library("Matrix")
library("BB")
library("matrixStats")
library("abind")
library("foreach")
library("doParallel")
library("numDeriv")
set.seed(12345)

####################
lSink <- TRUE # Saving output to *.txt file?
lParallel <- TRUE # Parallel Monte Carlo?
lSaveMCMatrix <- TRUE # Save Monte Carlo matrix?
lJacobian <- TRUE # Compute the jacobian of the best-response mapping?
lPopBR <- FALSE # Use population best-response mapping?
sSelectDGP <- "i" # Choose from PSD08's equilibria: "i", "ii", "iii"
vGridThetaRN <- seq(0,5,0.2)
sSelectFirstStep <- "freq" # Choose from: "freq", "kernel" or "true"
dSEqDGP <- 500 # Number of starting values to find different equilibria in the population.
dM <- 100 # Number of markets.
dK <- 500 # Number of iterations used in the NPL algorithm.
dStartVals <- 100 # Number of initial values used in algorithms.
dMC <- 500 # Number of Monte Carlo replications.
dFixAlpha <- 0.05 # Value of fixed alpha used in the relaxation method.
dCCPTol <- 10^(-5) # Tolerance used to assess convergence in the CCP space.
dNumEqTol <- 10^(-4) # Tolerance used to assess if two fixed points are different.
dCompTol <- 10^(-2) # Tolerance used to declare estimates as being the same.
####################


## Functions to be used.

# Compute a given player's expected F0, F1, and Z over all competitors' possible decisions.
fFZPANm1 <- function(MCCP, DI){ 
  a <-1
  MSUMANM1F0INIT <- matrix(rep(0, dNumStates^2), nrow=dNumStates, ncol=dNumStates)
  MSUMANM1F1INIT <- matrix(rep(0, dNumStates^2), nrow=dNumStates, ncol=dNumStates)
  MSUMANM1Z0INIT <- matrix(rep(0, dNumStates*length(vTheta)), nrow=dNumStates, ncol=length(vTheta))
  MSUMANM1Z1INIT <- matrix(rep(0, dNumStates*length(vTheta)), nrow=dNumStates, ncol=length(vTheta))
  
  VN <- 1:dN
  VCOLSELMI <- t(VN[-DI])
  VCOLSELI <- t(VN[DI])
  
  while (a <= dNumAlt^(dN-1)){
    VCCPANM1 <- rowProds(t(mANm1[a,]*t(MCCP[,VCOLSELMI])) + t((1-mANm1[a,])*t(1-MCCP[,VCOLSELMI])))
    MZ0ANM1 <- cbind(matrix(rep(0, dNumStates*(length(vTheta)-1)), nrow=dNumStates, ncol=length(vTheta)-1), mFullX[,DI])
    MZ1ANM1 <- cbind(rep(1,dNumStates), sum(mANm1[a,])*rep(1,dNumStates) , (1-mFullX[,DI]) , rep(0, dNumStates))
    
    MSUMANM1Z0 <- MSUMANM1Z0INIT + VCCPANM1*MZ0ANM1
    MSUMANM1Z0INIT <- MSUMANM1Z0
    MSUMANM1Z1 <- MSUMANM1Z1INIT + VCCPANM1*MZ1ANM1
    MSUMANM1Z1INIT <- MSUMANM1Z1
    
    MSUMANM1F1 <- MSUMANM1F1INIT + VCCPANM1*t(((colSums(t(mFullX[, VCOLSELMI])==mANm1[a,])==(dN-1))*(mFullX[, VCOLSELI]==1))*t(mFStates) ) 
    MSUMANM1F1INIT <- MSUMANM1F1
    
    MSUMANM1F0 <- MSUMANM1F0INIT + VCCPANM1*t(((colSums(t(mFullX[, VCOLSELMI])==mANm1[a,])==(dN-1))*(mFullX[, VCOLSELI]==0))*t(mFStates) ) 
    MSUMANM1F0INIT <- MSUMANM1F0
    
    a <- a+1
  }
  MF0PANM1 <- MSUMANM1F0
  MF1PANM1 <- MSUMANM1F1
  MZ0PANM1 <- MSUMANM1Z0
  MZ1PANM1 <- MSUMANM1Z1

  OUTFZPANM1 <- list("F0PANm1"=MF0PANM1, "F1PANm1"=MF1PANM1, "Z0PANm1"=MZ0PANM1, "Z1PANm1"=MZ1PANM1)
  return(OUTFZPANM1)
}


# Compute a given player's expected Z and E over all possible decisions.
fZEPAN <- function(MCCP, DI){ 
  b <-1
  MSUMANZ1INIT <- matrix(rep(0, dNumStates*length(vTheta)), nrow=dNumStates, ncol=length(vTheta))
  VSUMANE1INIT <- rep(0, dNumStates)
              
  while (b <= dNumAlt^dN){
    VCCPAN <- rowProds(t(mAN[b,]*t(MCCP)) + t((1-mAN[b,])*t(1-MCCP)))

    MZ1AN <- mAN[b,DI]*(cbind(rep(1,dNumStates), sum(mAN[b,3-DI])*rep(1,dNumStates) , (1-mFullX[,DI]) , rep(0, dNumStates))) + (1-mAN[b,DI])*(cbind(matrix(rep(0, dNumStates*(length(vTheta)-1)), nrow=dNumStates, ncol=length(vTheta)-1), mFullX[,DI]))
    MSUMANZ1 <- MSUMANZ1INIT + VCCPAN*MZ1AN 
    MSUMANZ1INIT <- MSUMANZ1
    
    VEAN <- (1/sqrt(1))*mAN[b,DI]*dnorm(qnorm(MCCP[,DI]))/MCCP[,DI]
    VSUMANE1 <- VSUMANE1INIT + VCCPAN*VEAN
    VSUMANE1INIT <- VSUMANE1
      
    b <- b+1
  }
  
  MZPAN <- MSUMANZ1
  MEPAN <- VSUMANE1

  OUTZEPAN <- list("ZPAN"=MZPAN, "EPAN"=MEPAN)
}


# Collect all players' expected F, Z and E.
fAllFZEP <- function(MCCPS){ 
  MALLF0PANM1 <- matrix(data=NA, nrow=dNumStates, ncol=dN*dNumStates)
  MALLF1PANM1 <- matrix(data=NA, nrow=dNumStates, ncol=dN*dNumStates)
  MALLZ0PANM1 <- matrix(data=NA, nrow=dNumStates, ncol=length(vTheta)*dN)
  MALLZ1PANM1 <- matrix(data=NA, nrow=dNumStates, ncol=length(vTheta)*dN)
  MALLZPAN <- matrix(data=NA, nrow=dNumStates, ncol=length(vTheta)*dN)
  MALLEPAN <- matrix(data=NA, nrow=dNumStates, ncol=dN)
  for (o in 1:dN){
    OUTFZPANM1 <- fFZPANm1(MCCPS, o)
    MALLF0PANM1[,((o-1)*dNumStates + 1):(o*dNumStates)] <- OUTFZPANM1$F0PANm1
    MALLF1PANM1[,((o-1)*dNumStates + 1):(o*dNumStates)] <- OUTFZPANM1$F1PANm1
    MALLZ0PANM1[,((o-1)*length(vTheta) + 1):(o*length(vTheta))] <- OUTFZPANM1$Z0PANm1
    MALLZ1PANM1[,((o-1)*length(vTheta) + 1):(o*length(vTheta))] <- OUTFZPANM1$Z1PANm1

    OUTZEPAN <- fZEPAN(MCCPS, o)
    MALLZPAN[,((o-1)*length(vTheta) + 1):(o*length(vTheta))] <- OUTZEPAN$ZPAN
    MALLEPAN[,o] <- OUTZEPAN$EPAN
  }

  OUTALLFZEP <- list("allF0PANm1"=MALLF0PANM1, "allF1PANm1"=MALLF1PANM1, "allZ0PANm1"=MALLZ0PANM1, "allZ1PANm1"=MALLZ1PANM1, "allZPAN"=MALLZPAN, "allEPAN"=MALLEPAN)
  return(OUTALLFZEP)  
}

# Compute expected F over all possible decisions (same for all players).
fFP <- function(MCCPS){ 
  b <-1
  MSUMANF1INIT <- matrix(rep(0, dNumStates^2), nrow=dNumStates, ncol=dNumStates)
  while (b <= dNumAlt^dN){
    VCCPAN <- rowProds(t(mAN[b,]*t(MCCPS)) + t((1-mAN[b,])*t(1-MCCPS)));
    MSUMANF1 <- MSUMANF1INIT + VCCPAN*t(((colSums(t(mFullX)==mAN[b,])==dN))*t(mFStates)) 
    MSUMANF1INIT <- MSUMANF1
    
    b <- b+1
  }  
  MFPAN <- MSUMANF1

  return(MFPAN)
}


# Construct best-response mapping.
fBRmap <- function(VTHETA, OUTALLFZEP, MINVERT){ 
  MF0PANM1 <- OUTALLFZEP$allF0PANm1
  MF1PANM1 <- OUTALLFZEP$allF1PANm1
  MZ0PANM1 <- OUTALLFZEP$allZ0PANm1
  MZ1PANM1 <- OUTALLFZEP$allZ1PANm1
  MZPAN <- OUTALLFZEP$allZPAN
  MEPAN <- OUTALLFZEP$allEPAN
  
  MPSY <- matrix(data=NA, nrow=dNumStates, ncol=dN)
  for (n in 1:dN){
    VVALPI <- MINVERT%*%(MZPAN[,((n-1)*length(VTHETA)+1):(n*length(VTHETA))]%*%VTHETA + MEPAN[,n])
    VCONDVALPI1 <- MZ1PANM1[,((n-1)*length(VTHETA)+1):(n*length(VTHETA))]%*%VTHETA + dBeta*MF1PANM1[,((n-1)*dNumStates+1):(n*dNumStates)]%*%VVALPI
    VCONDVALPI0 <- MZ0PANM1[,((n-1)*length(VTHETA)+1):(n*length(VTHETA))]%*%VTHETA + dBeta*MF0PANM1[,((n-1)*dNumStates+1):(n*dNumStates)]%*%VVALPI
    MPSY[,n] <- pnorm(VCONDVALPI1-VCONDVALPI0, mean=0, sd=sqrt(1))
  }

  return(MPSY)
}


# Find a root of the best response mapping.
# i: VINITCCPS <- 1-c(0.27, 0.39, 0.20, 0.25, 0.72, 0.58, 0.78, 0.71)
# ii: VINITCCPS <- 1-c(0.38, 0.69, 0.17, 0.39, 0.47, 0.16, 0.7, 0.42)
# iii: VINITCCPS <- 1-c(0.42, 0.70, 0.16, 0.41, 0.42, 0.16, 0.70, 0.41)
# iv: VINITCCPS <- 1-c(0.72, 0.78, 0.58, 0.71, 0.27, 0.2, 0.39, 0.25)
# v: VINITCCPS <- 1-c(0.47, 0.7, 0.16, 0.42, 0.38, 0.17, 0.69, 0.39)

fRootDGPCCPs <- function(VCCPS){ 
  MCCPS <- matrix(data = VCCPS, nrow=dNumStates, ncol=dN)

  OUTALLFZEP <- fAllFZEP(MCCPS)   
  MFPAN <- fFP(MCCPS)  
  MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)

  MCCPBR <- fBRmap(vTheta, OUTALLFZEP, MINVERT)
  
  c(MCCPS - MCCPBR)
}

# Solve CCPs using multiple starting values.
fMultiSolveDGPCCPs <- function(DSEQDGP){ 
  ATRUECCPS <- array(data=NA, dim=c(dNumStates,dN,1))
  DSUCCESSCONVTRUE <- 0
  for (s in 1:DSEQDGP){
    VINITCCPS <- runif(dNumStates*dN)
    OUTSOLUTIONTRUE <- BBsolve(VINITCCPS, fRootDGPCCPs, method=c(2,3,1), control=list(maxit=dK), quiet=TRUE)
    
    if (OUTSOLUTIONTRUE$convergence==0){
      DSUCCESSCONVTRUE <- DSUCCESSCONVTRUE + (OUTSOLUTIONTRUE$convergence==0)
      MTRUECCPS <- matrix(c(OUTSOLUTIONTRUE$par), nrow=dNumStates, ncol=dN)  
      if (DSUCCESSCONVTRUE==1){
        ATRUECCPS[,,1] <- MTRUECCPS
      } else {
        VCHECKEQ <- c()
        for (t in 1:dim(ATRUECCPS)[3]){
          VCHECKEQ[t] <- max(abs(MTRUECCPS-ATRUECCPS[,,t]))
        }  
        if (min(VCHECKEQ)>dNumEqTol){
          ATRUECCPS <- abind(ATRUECCPS, MTRUECCPS, along=3)
        }
      }
    }
  }

  if (dim(ATRUECCPS)[3]>1){
    VDISTEQ <- c()
    for (t in 1:dim(ATRUECCPS)[3]){
      VDISTEQ[t] <- max(abs(ATRUECCPS[,,1]-ATRUECCPS[,,t]))
    } 
    DDISTEQ <- max(VDISTEQ)    
  } else{
    DDISTEQ <- 0
  }


  OUTSOLVEDGPCCPS <- list("trueccps"=ATRUECCPS, "success"=DSUCCESSCONVTRUE, "disteq"=DDISTEQ)
  return(OUTSOLVEDGPCCPS)
}


# Construct the NPL mapping.
fNPLmap <- function(VCCPS){ 
  MCCPS <- matrix(data = VCCPS, nrow=dNumStates, ncol=dN)
  MCCPS <- (MCCPS>0)*(MCCPS<1)*MCCPS + (MCCPS>=1)*(1-10^(-10)) + (MCCPS<=0)*10^(-10)

  OUTALLFZEP <- fAllFZEP(MCCPS)
  MFPAN <- fFP(MCCPS)    
  MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
  
  OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZEP, MINVERT=MINVERT)
  VNPLTHETA <- OUTMAXLOGLIK$est
  MNPLCCPS <- fBRmap(OUTMAXLOGLIK$est, OUTALLFZEP, MINVERT)
  DNPLLOGLIK <- OUTMAXLOGLIK$min

  OUTNPLMAP <- list("NPLtheta"=VNPLTHETA, "NPLccps"=MNPLCCPS, "NPLloglik"=DNPLLOGLIK)
  return(OUTNPLMAP)
}

# Find a root of the NPL mapping.
fPhiNPL <- function(VCCPS){ 
  MCCPS <- matrix(data = VCCPS, nrow=dNumStates, ncol=dN)
  MNPLCCPS <- fNPLmap(VCCPS)$NPLccps
  c(MCCPS - MNPLCCPS)
}

# Log-likelihood function.
fLogLik <- function(VTHETA, OUTALLFZEP, MINVERT){ 
  MPSY <- fBRmap(VTHETA, OUTALLFZEP, MINVERT)
  MPSY <- (MPSY>0)*(MPSY<1)*MPSY + (MPSY>=1)*(1-10^(-10)) + (MPSY<=0)*10^(-10)

  MINDIVLOGLIK <- matrix(data=NA, nrow=(lPopBR==TRUE)*(nrow(mAN)*dNumStates) + (lPopBR==FALSE)*dM, ncol=dN)
  for (n in 1:dN){
    VPSY <- MPSY[,n]
    if (lPopBR==TRUE){
      VENUMSTATE <- kronecker(seq(1,dNumStates), rep(1,nrow(mAN)))
      MA1 <- kronecker(rep(1,dNumStates), mAN)
      MINDIVLOGLIK[,n] <- (mTrueCCPs[VENUMSTATE,n]*MA1[,n]*log(VPSY[VENUMSTATE]) + (1-mTrueCCPs[VENUMSTATE,n])*(1-MA1[,n])*log(1-VPSY[VENUMSTATE]))*vSSX[VENUMSTATE]
    } else{
      MA1 <- mA1
      MINDIVLOGLIK[,n] <- MA1[,n]*log(VPSY[vIndState]) + (1-MA1[,n])*log(1-VPSY[vIndState])
    }
  }
  -(1/(nrow(MA1)*dN))*sum(MINDIVLOGLIK) 
}

# Function used to compute the Jacobian.
fJacobian <- function(VCCPS){ 
  c(fNPLmap(VCCPS)$NPLccps)
}

# 2-step estimator with mCCPhat in first step.
f2step <- function(MCCPHAT){ 
  OUTALLFZEP <- fAllFZEP(MCCPHAT)
  MFPAN <- fFP(MCCPHAT)    
  MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      
  OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZEP, MINVERT=MINVERT)
  if (OUTMAXLOGLIK$code <=3){
    VTHETAEST <- OUTMAXLOGLIK$est
    DLOGLIKEST <- OUTMAXLOGLIK$min
  } else{
    cat("Maximum likelihood estimation failed to converge in two-step estimator. \n")
    VTHETAEST <- rep(NA, length(vTheta))
    DLOGLIKEST <- NA
  }
  
  OUT2STEP <- list("thetaest"=VTHETAEST, "loglikest"=DLOGLIKEST) 
  return(OUT2STEP)
}

# NPL algorigthm.
fNPLAlgo <- function(MCCPS, DK){ 
  DDIFFCCPS <- 10
  k <- 0
  VTIMEITER <- c()
  VTIMEUPTHETA <- c()
  VTIMEUPCCP <- c()

  while(k <= DK-1 & DDIFFCCPS > dCCPTol){    
    DSTARTITER <- Sys.time()
    DSTARTTHETA <- Sys.time()

    k <- k+1
    OUTALLFZEP <- fAllFZEP(MCCPS)
    MFPAN <- fFP(MCCPS)    
    MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
        
    OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZEP, MINVERT=MINVERT)
    if (OUTMAXLOGLIK$code <=3){
      VTHETAEST <- OUTMAXLOGLIK$est
      DLOGLIKEST <- OUTMAXLOGLIK$min
    } else{
      cat("Maximum likelihood estimation failed to converge at k = ", k, " in NPL algorithm. \n")
      VTHETAEST <- rep(NA, length(vTheta))
      DLOGLIKEST <- NA
    }
    
    DENDTHETA <- Sys.time()
    DSTARTCCP <- Sys.time()

    MNEWCCPS <- fBRmap(OUTMAXLOGLIK$est, OUTALLFZEP, MINVERT)
    MNEWCCPS <- (MNEWCCPS>0)*(MNEWCCPS<1)*MNEWCCPS + (MNEWCCPS>=1)*(1-10^(-10)) + (MNEWCCPS<=0)*10^(-10)

    DENDCCP <- Sys.time()

    DDIFFCCPS <- max(abs(MNEWCCPS - MCCPS))
    MCCPS <- MNEWCCPS

    DENDITER <- Sys.time()    
    
    VTIMEITER <- c(VTIMEITER, difftime(DENDITER, DSTARTITER, units="secs"))
    VTIMEUPTHETA <- c(VTIMEUPTHETA, difftime(DENDTHETA, DSTARTTHETA, units="secs"))
    VTIMEUPCCP <- c(VTIMEUPCCP, difftime(DENDCCP, DSTARTCCP, units="secs"))
  }

  OUTNPLALGO <- list("thetaest"=VTHETAEST, "ccpsest"=c(MCCPS), "loglikest"=DLOGLIKEST, "diffccps"=DDIFFCCPS, "k"=k, "avgtimes" = c(mean(VTIMEITER), mean(VTIMEUPTHETA), mean(VTIMEUPCCP)), "sdtimes" = c(sd(VTIMEITER), sd(VTIMEUPTHETA), sd(VTIMEUPCCP))) 
  return(OUTNPLALGO)
}

# Perform NPL algorithm using multiple starting values.
fMultiNPLAlgo <- function(DK, DSTARTVALS, MCCPHAT){ 
  DSUCCESS <- 0
  AMULTIFXPT <- array(data=NA, dim=c(dNumStates,dN,1))
  DMULTIFXPT <- 0
  VMULTILOGLIK <- Inf

  for (s in 1:DSTARTVALS){
    if (s==1){
      VINITCCPS <- c(MCCPHAT)
    } else if (s==2){
      OUT2STEP <- f2step(MCCPHAT)
      VTHETA2STEP <- OUT2STEP$thetaest
      OUTALLFZEP <- fAllFZEP(MCCPHAT)
      MFPAN <- fFP(MCCPHAT)    
      MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      MBRCCP <- fBRmap(VTHETA2STEP, OUTALLFZEP, MINVERT)
      VINITCCPS <- c(MBRCCP)
    } else {
      set.seed(1234+s*1000)
      VINITCCPS <- runif(dNumStates*dN, min=0, max=1)
    }  
    MINITCCPS <- matrix(VINITCCPS, nrow=dNumStates, ncol=dN)
    OUTNPLALGO <- fNPLAlgo(MINITCCPS, DK)
    if (!is.na(OUTNPLALGO$loglikest)){
      DSUCCESS <- DSUCCESS + (!is.na(OUTNPLALGO$loglikest))
      VCCPSLASTCHECK <- OUTNPLALGO$ccpsest
      VTHETALASTCHECK <- OUTNPLALGO$thetaest
      DLOGLIKLASTCHECK <- OUTNPLALGO$loglikest
      DDIFFCCPSLASTCHECK <- OUTNPLALGO$diffccps
      DKLASTCHECK <- OUTNPLALGO$k
      VAVGTIMECHECK <- OUTNPLALGO$avgtimes
      VSDTIMECHECK <- OUTNPLALGO$sdtimes
      if (DSUCCESS==1){
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK < DK){        
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK        
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK == DK & DKLAST == DK){        
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK        
      } 
      DISFXPT <- DDIFFCCPSLASTCHECK<dCCPTol
      if (DISFXPT==TRUE){
        MMULTIFXPT <- matrix(VCCPSLASTCHECK, nrow=dNumStates, ncol=dN)
        if (DMULTIFXPT==0){
          DMULTIFXPT <- DMULTIFXPT + DISFXPT
          AMULTIFXPT[,,1] <- MMULTIFXPT
          VMULTILOGLIK <- DLOGLIKLASTCHECK
        } else{
          VCHECKFXPT <- c()
          for (t in 1:dim(AMULTIFXPT)[3]){
            VCHECKFXPT[t] <- max(abs(MMULTIFXPT-AMULTIFXPT[,,t]))
          }  
          if (min(VCHECKFXPT)>dNumEqTol){
            AMULTIFXPT <- abind(AMULTIFXPT, MMULTIFXPT, along=3)
            DMULTIFXPT <- DMULTIFXPT + DISFXPT
            VMULTILOGLIK <- c(VMULTILOGLIK, DLOGLIKLASTCHECK)
          }
        }
      }    
    }       
  }

  if (dim(AMULTIFXPT)[3] > 1){
    AMULTIFXPT <- AMULTIFXPT[,,order(VMULTILOGLIK)]
    VDISTFXPT <- c()
    for (t in 1:dim(AMULTIFXPT)[3]){
      VDISTFXPT[t] <- max(abs(AMULTIFXPT[,,1]-AMULTIFXPT[,,t]))
    } 
    DDISTFXPT <- max(VDISTFXPT)
  } else{
    DDISTFXPT <- 0
  }

  OUTMULTINPLALGO <- list("thetalast"=VTHETALAST, "ccplast"=VCCPSLAST, "logliklast"=DLOGLIKLAST, "diffccpslast"=DDIFFCCPSLAST, "k"=DKLAST, "numbfxpt"=DMULTIFXPT, "distfxpt"=DDISTFXPT, "avgtimes"=VAVGTIME, "sdtimes"=VSDTIME)
  return(OUTMULTINPLALGO)
}

# Relaxation NPL algorigthm.
fRelaxAlgo <- function(MCCPS, DK, DFIXALPHA){ 
  DDIFFCCPS <- 10
  k <- 0
  VTIMEITER <- c()
  VTIMEUPTHETA <- c()
  VTIMEUPCCP <- c()
   
  while(k <= DK-1 & DDIFFCCPS > dCCPTol){  
    DSTARTITER <- Sys.time()
    DSTARTTHETA <- Sys.time()

    k <- k+1
    OUTALLFZEP <- fAllFZEP(MCCPS)
    MFPAN <- fFP(MCCPS)    
    MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
        
    OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZEP, MINVERT=MINVERT)
    if (OUTMAXLOGLIK$code <=3){
      VTHETAEST <- OUTMAXLOGLIK$est
      DLOGLIKEST <- OUTMAXLOGLIK$min
    } else{
      cat("Maximum likelihood estimation failed to converge at k = ", k, " in relaxation algorithm. \n")
      VTHETAEST <- rep(NA, length(vTheta))
      DLOGLIKEST <- NA
    }

    DENDTHETA <- Sys.time()
    DSTARTCCP <- Sys.time()
    
    MBRCCPS <- fBRmap(OUTMAXLOGLIK$est, OUTALLFZEP, MINVERT)
    MNEWCCPS <- (MBRCCPS^DFIXALPHA)*(MCCPS^(1-DFIXALPHA))
    MNEWCCPS <- (MNEWCCPS>0)*(MNEWCCPS<1)*MNEWCCPS + (MNEWCCPS>=1)*(1-10^(-10)) + (MNEWCCPS<=0)*10^(-10)

    DENDCCP <- Sys.time()

    DDIFFCCPS <- max(abs(MNEWCCPS - MCCPS))
    MCCPS <- MNEWCCPS

    DENDITER <- Sys.time()
    
    VTIMEITER <- c(VTIMEITER, difftime(DENDITER, DSTARTITER, units="secs"))
    VTIMEUPTHETA <- c(VTIMEUPTHETA, difftime(DENDTHETA, DSTARTTHETA, units="secs"))
    VTIMEUPCCP <- c(VTIMEUPCCP, difftime(DENDCCP, DSTARTCCP, units="secs"))    
  }

  OUTRELAXALGO <- list("thetaest"=VTHETAEST, "ccpsest"=c(MCCPS), "loglikest"=DLOGLIKEST, "diffccps"=DDIFFCCPS, "k"=k, "avgtimes" = c(mean(VTIMEITER), mean(VTIMEUPTHETA), mean(VTIMEUPCCP)), "sdtimes" = c(sd(VTIMEITER), sd(VTIMEUPTHETA), sd(VTIMEUPCCP))) 
  return(OUTRELAXALGO)
}

# Perform relaxation method using multiple starting values.
fMultiRelaxAlgo <- function(DK, DSTARTVALS, MCCPHAT, DFIXALPHA){ 
  DSUCCESS <- 0
  AMULTIFXPT <- array(data=NA, dim=c(dNumStates,dN,1))
  DMULTIFXPT <- 0
  VMULTILOGLIK <- Inf

  for (s in 1:DSTARTVALS){
    if (s==1){
      VINITCCPS <- c(MCCPHAT)
    } else if (s==2){
      OUT2STEP <- f2step(MCCPHAT)
      VTHETA2STEP <- OUT2STEP$thetaest
      OUTALLFZEP <- fAllFZEP(MCCPHAT)
      MFPAN <- fFP(MCCPHAT)    
      MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      MBRCCP <- fBRmap(VTHETA2STEP, OUTALLFZEP, MINVERT)
      VINITCCPS <- c(MBRCCP)
    } else{
      set.seed(1234+s*1000)
      VINITCCPS <- runif(dNumStates*dN, min=0, max=1)
    }  
    MINITCCPS <- matrix(VINITCCPS, nrow=dNumStates, ncol=dN)
    OUTRELAXALGO <- fRelaxAlgo(MINITCCPS, DK, DFIXALPHA)
    if (!is.na(OUTRELAXALGO$loglikest)){    
      DSUCCESS <- DSUCCESS + (!is.na(OUTRELAXALGO$loglikest))
      VCCPSLASTCHECK <- OUTRELAXALGO$ccpsest
      VTHETALASTCHECK <- OUTRELAXALGO$thetaest
      DLOGLIKLASTCHECK <- OUTRELAXALGO$loglikest
      DDIFFCCPSLASTCHECK <- OUTRELAXALGO$diffccps
      DKLASTCHECK <- OUTRELAXALGO$k
      VAVGTIMECHECK <- OUTRELAXALGO$avgtimes
      VSDTIMECHECK <- OUTRELAXALGO$sdtimes
      if (DSUCCESS==1){
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK < DK){        
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK == DK & DKLAST == DK){        
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK 
      } 

      DISFXPT <- DDIFFCCPSLASTCHECK<dCCPTol
      if (DISFXPT==TRUE){
        MMULTIFXPT <- matrix(VCCPSLASTCHECK, nrow=dNumStates, ncol=dN)
        if (DMULTIFXPT==0){
          DMULTIFXPT <- DMULTIFXPT + DISFXPT
          AMULTIFXPT[,,1] <- MMULTIFXPT
          VMULTILOGLIK <- DLOGLIKLASTCHECK          
        } else{
          VCHECKFXPT <- c()
          for (t in 1:dim(AMULTIFXPT)[3]){
            VCHECKFXPT[t] <- max(abs(MMULTIFXPT-AMULTIFXPT[,,t]))
          }  
          if (min(VCHECKFXPT)>dNumEqTol){
            AMULTIFXPT <- abind(AMULTIFXPT, MMULTIFXPT, along=3)
            DMULTIFXPT <- DMULTIFXPT + DISFXPT
            VMULTILOGLIK <- c(VMULTILOGLIK, DLOGLIKLASTCHECK)            
          }
        }
      }    
    }       
  }

  if (dim(AMULTIFXPT)[3] > 1){
    AMULTIFXPT <- AMULTIFXPT[,,order(VMULTILOGLIK)]
    VDISTFXPT <- c()
    for (t in 1:dim(AMULTIFXPT)[3]){
      VDISTFXPT[t] <- max(abs(AMULTIFXPT[,,1]-AMULTIFXPT[,,t]))
    } 
    DDISTFXPT <- max(VDISTFXPT)
  } else{
    DDISTFXPT <- 0
  }

  OUTMULTIRELAXALGO <- list("thetalast"=VTHETALAST, "ccplast"=VCCPSLAST, "logliklast"=DLOGLIKLAST, "diffccpslast"=DDIFFCCPSLAST, "k"=DKLAST, "numbfxpt"=DMULTIFXPT, "distfxpt"=DDISTFXPT, "avgtimes"=VAVGTIME, "sdtimes"=VSDTIME)
  return(OUTMULTIRELAXALGO)
}

# Spectral algorigthm.
fSpecAlgo <- function(MCCPS, DK){ 
  k <- 1
  VTIMEITER <- c()
  VTIMEUPTHETA <- c()
  VTIMEUPCCP <- c()

  DSTARTITER <- Sys.time()
  DSTARTTHETA <- Sys.time()

  OUTALLFZEP <- fAllFZEP(MCCPS)
  MFPAN <- fFP(MCCPS)    
  MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      
  OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZEP, MINVERT=MINVERT)
  if (OUTMAXLOGLIK$code <=3){
    VTHETAEST <- OUTMAXLOGLIK$est
    DLOGLIKEST <- OUTMAXLOGLIK$min
  } else{
    cat("Maximum likelihood estimation failed to converge at k = ", k, " in spectral algorithm. \n")
    VTHETAEST <- rep(NA, length(vTheta))
    DLOGLIKEST <- NA
  }

  DENDTHETA <- Sys.time()
  DSTARTCCP <- Sys.time()

  VPHI <- c(MCCPS) - c(fBRmap(OUTMAXLOGLIK$est, OUTALLFZEP, MINVERT))
  DALPHA <- min(1,1/sqrt(sum(VPHI^2)))

  VNEWCCPS <- c(MCCPS) - DALPHA*VPHI
  VNEWCCPS <- (VNEWCCPS>0)*(VNEWCCPS<1)*VNEWCCPS + (VNEWCCPS>=1)*(1-10^(-10)) + (VNEWCCPS<=0)*10^(-10)
  
  DENDCCP <- Sys.time()

  MNEWCCPS <- matrix(VNEWCCPS, nrow=dNumStates, ncol=dN)
  DDIFFCCPS <- max(abs(MNEWCCPS - MCCPS))
  
  DENDITER <- Sys.time()

  VTIMEITER <- difftime(DENDITER, DSTARTITER, units="secs")
  VTIMEUPTHETA <- difftime(DENDTHETA, DSTARTTHETA, units="secs")
  VTIMEUPCCP <- difftime(DENDCCP, DSTARTCCP, units="secs")

  while(k <= DK-1 & DDIFFCCPS > dCCPTol){    
    DSTARTITER <- Sys.time()
    DSTARTTHETA <- Sys.time()

    k <- k+1 
    
    OUTALLFZENEWP <- fAllFZEP(MNEWCCPS)
    MFNEWPAN <- fFP(MNEWCCPS)    
    MNEWINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFNEWPAN)

    OUTMAXLOGLIK <- nlm(fLogLik, vTheta, OUTALLFZEP=OUTALLFZENEWP, MINVERT=MNEWINVERT)
    if (OUTMAXLOGLIK$code <=3){
      VTHETAEST <- OUTMAXLOGLIK$est
      DLOGLIKEST <- OUTMAXLOGLIK$min
    } else{
      cat("Maximum likelihood estimation failed to converge at k = ", k, " in spectral algorithm. \n")
      VTHETAEST <- OUTMAXLOGLIK$est
      DLOGLIKEST <- NA
    }

    DENDTHETA <- Sys.time()
    DSTARTCCP <- Sys.time()

    VNEWPHI <- c(MNEWCCPS) - c(fBRmap(OUTMAXLOGLIK$est, OUTALLFZENEWP, MNEWINVERT))
    DNEWALPHA <- c((t(VNEWCCPS-c(MCCPS))%*%(VNEWCCPS-c(MCCPS)))/(t(VNEWCCPS-c(MCCPS))%*%(VNEWPHI-VPHI))  )

    MCCPS <- matrix(VNEWCCPS, nrow=dNumStates, ncol=dN)
    VPHI <- VNEWPHI
    DALPHA <- DNEWALPHA

    VNEWCCPS <- c(MCCPS) - DALPHA*VPHI
    VNEWCCPS <- (VNEWCCPS>0)*(VNEWCCPS<1)*VNEWCCPS + (VNEWCCPS>=1)*(1-10^(-10)) + (VNEWCCPS<=0)*10^(-10)

    DENDCCP <- Sys.time()

    MNEWCCPS <- matrix(VNEWCCPS, nrow=dNumStates, ncol=dN)
    DDIFFCCPS <- max(abs(MNEWCCPS - MCCPS))   


    DENDITER <- Sys.time()
    
    VTIMEITER <- c(VTIMEITER, difftime(DENDITER, DSTARTITER, units="secs"))
    VTIMEUPTHETA <- c(VTIMEUPTHETA, difftime(DENDTHETA, DSTARTTHETA, units="secs"))
    VTIMEUPCCP <- c(VTIMEUPCCP, difftime(DENDCCP, DSTARTCCP, units="secs"))    
  }

  OUTSPECALGO <- list("thetaest"=VTHETAEST, "ccpsest"=c(MCCPS), "loglikest"=DLOGLIKEST, "diffccps"=DDIFFCCPS, "k"=k, "avgtimes" = c(mean(VTIMEITER), mean(VTIMEUPTHETA), mean(VTIMEUPCCP)), "sdtimes" = c(sd(VTIMEITER), sd(VTIMEUPTHETA), sd(VTIMEUPCCP)))
  return(OUTSPECALGO)
}

# Perform spectral algorithm using multiple starting values.
fMultiSpecAlgo <- function(DK, DSTARTVALS, MCCPHAT){ 
  DSUCCESS <- 0
  AMULTIFXPT <- array(data=NA, dim=c(dNumStates,dN,1))
  DMULTIFXPT <- 0
  VMULTILOGLIK <- Inf

  for (s in 1:DSTARTVALS){
    if (s==1){
      VINITCCPS <- c(MCCPHAT)
    } else if (s==2){
      OUT2STEP <- f2step(MCCPHAT)
      VTHETA2STEP <- OUT2STEP$thetaest
      OUTALLFZEP <- fAllFZEP(MCCPHAT)
      MFPAN <- fFP(MCCPHAT)    
      MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      MBRCCP <- fBRmap(VTHETA2STEP, OUTALLFZEP, MINVERT)
      VINITCCPS <- c(MBRCCP)
    } else{
      set.seed(1234+s*1000)
      VINITCCPS <- runif(dNumStates*dN, min=0, max=1)
    }  
    MINITCCPS <- matrix(VINITCCPS, nrow=dNumStates, ncol=dN)
    OUTSPECALGO <- fSpecAlgo(MINITCCPS, DK)
    if (!is.na(OUTSPECALGO$loglikest)){
      DSUCCESS <- DSUCCESS + (!is.na(OUTSPECALGO$loglikest))
      VCCPSLASTCHECK <- OUTSPECALGO$ccpsest
      VTHETALASTCHECK <- OUTSPECALGO$thetaest
      DLOGLIKLASTCHECK <- OUTSPECALGO$loglikest
      DDIFFCCPSLASTCHECK <- OUTSPECALGO$diffccps
      DKLASTCHECK <- OUTSPECALGO$k
      VAVGTIMECHECK <- OUTSPECALGO$avgtimes
      VSDTIMECHECK <- OUTSPECALGO$sdtimes
      if (DSUCCESS==1){
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK        
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK < DK){                
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK        
      } else if(DLOGLIKLASTCHECK <= DLOGLIKLAST & DKLASTCHECK == DK & DKLAST == DK){        
        VCCPSLAST <- VCCPSLASTCHECK
        VTHETALAST <- VTHETALASTCHECK
        DLOGLIKLAST <- DLOGLIKLASTCHECK
        DDIFFCCPSLAST <- DDIFFCCPSLASTCHECK
        DKLAST <- DKLASTCHECK
        VAVGTIME <- VAVGTIMECHECK
        VSDTIME <- VSDTIMECHECK        
      } 
      DISFXPT <- DDIFFCCPSLASTCHECK<dCCPTol
      if (DISFXPT==TRUE){
        MMULTIFXPT <- matrix(VCCPSLASTCHECK, nrow=dNumStates, ncol=dN)
        if (DMULTIFXPT==0){
          DMULTIFXPT <- DMULTIFXPT + DISFXPT
          AMULTIFXPT[,,1] <- MMULTIFXPT
          VMULTILOGLIK <- DLOGLIKLASTCHECK
        } else{
          VCHECKFXPT <- c()
          for (t in 1:dim(AMULTIFXPT)[3]){
            VCHECKFXPT[t] <- max(abs(MMULTIFXPT-AMULTIFXPT[,,t]))
          }  
          if (min(VCHECKFXPT)>dNumEqTol){
            AMULTIFXPT <- abind(AMULTIFXPT, MMULTIFXPT, along=3)
            DMULTIFXPT <- DMULTIFXPT + DISFXPT
            VMULTILOGLIK <- c(VMULTILOGLIK, DLOGLIKLASTCHECK)
          }
        }
      }               
    }       
  }

  if (dim(AMULTIFXPT)[3] > 1){
    AMULTIFXPT <- AMULTIFXPT[,,order(VMULTILOGLIK)]
    VDISTFXPT <- c()
    for (t in 1:dim(AMULTIFXPT)[3]){
      VDISTFXPT[t] <- max(abs(AMULTIFXPT[,,1]-AMULTIFXPT[,,t]))
    } 
    DDISTFXPT <- max(VDISTFXPT)
  } else{
    DDISTFXPT <- 0
  }

  OUTMULTISPECALGO <- list("thetalast"=VTHETALAST, "ccplast"=VCCPSLAST, "logliklast"=DLOGLIKLAST, "diffccpslast"=DDIFFCCPSLAST, "k"=DKLAST, "numbfxpt"=DMULTIFXPT, "distfxpt"=DDISTFXPT, "avgtimes"=VAVGTIME, "sdtimes"=VSDTIME)
  return(OUTMULTISPECALGO)
}

# Compute NPL estimator using multiple starting values.
fMultiSolveNPL <- function(DSTARTVALS, MCCPHAT){ 
  DSUCCESSCONVTRUE <- 0
  AMULTIFXPT <- array(data=NA, dim=c(dNumStates,dN,1))
  DMULTIFXPT <- 0
  VMULTILOGLIK <- Inf
  
  for (s in 1:DSTARTVALS){
    if (s==1){
      VINITCCPS <- c(MCCPHAT)
    } else if (s==2){
      OUT2STEP <- f2step(MCCPHAT)
      VTHETA2STEP <- OUT2STEP$thetaest
      OUTALLFZEP <- fAllFZEP(MCCPHAT)
      MFPAN <- fFP(MCCPHAT)    
      MINVERT <- solve(diag(rep(dNumStates)) - dBeta*MFPAN)
      MBRCCP <- fBRmap(VTHETA2STEP, OUTALLFZEP, MINVERT)
      VINITCCPS <- c(MBRCCP)
    } else{
      set.seed(1234+s*1000)
      VINITCCPS <- runif(dNumStates*dN, min=0, max=1)
    }

    DSTARTFULL <- Sys.time()
    OUTSOLUTIONNPL <- BBsolve(VINITCCPS, fPhiNPL, method=c(2,3,1), control=list(maxit=dK), quiet=TRUE)
    DENDFULL <- Sys.time()
    DTIMEFULL <- difftime(DENDFULL, DSTARTFULL, units="secs")

    if (OUTSOLUTIONNPL$convergence==0){
      DSUCCESSCONVTRUE <- DSUCCESSCONVTRUE + (OUTSOLUTIONNPL$convergence==0)
      VNPLCCPSCHECK <- OUTSOLUTIONNPL$par
      OUTNPLMAP <- fNPLmap(VNPLCCPSCHECK)
      VNPLTHETACHECK <- OUTNPLMAP$NPLtheta
      DNPLLOGLIKCHECK <- OUTNPLMAP$NPLloglik
      DNPLTIMEFULLCHECK <- DTIMEFULL
      if (DSUCCESSCONVTRUE==1){
        VNPLCCPS <- VNPLCCPSCHECK
        VNPLTHETA <- VNPLTHETACHECK
        DNPLLOGLIK <- DNPLLOGLIKCHECK
        DNPLTIMEFULL <- DNPLTIMEFULLCHECK
      } else if(DNPLLOGLIKCHECK < DNPLLOGLIK){
        VNPLCCPS <- VNPLCCPSCHECK
        VNPLTHETA <- VNPLTHETACHECK
        DNPLLOGLIK <- DNPLLOGLIKCHECK
        DNPLTIMEFULL <- DNPLTIMEFULLCHECK
      }

      DISFXPT <- max(abs(fPhiNPL(VNPLCCPSCHECK)))<dCCPTol
      if (DISFXPT==TRUE){
        MMULTIFXPT <- matrix(VNPLCCPSCHECK, nrow=dNumStates, ncol=dN)
        if (DMULTIFXPT==0){
          DMULTIFXPT <- DMULTIFXPT + DISFXPT
          AMULTIFXPT[,,1] <- MMULTIFXPT
          VMULTILOGLIK <- DNPLLOGLIK
        } else{
          VCHECKFXPT <- c()
          for (t in 1:dim(AMULTIFXPT)[3]){
            VCHECKFXPT[t] <- max(abs(MMULTIFXPT-AMULTIFXPT[,,t]))
          }  
          if (min(VCHECKFXPT)>dNumEqTol){
            AMULTIFXPT <- abind(AMULTIFXPT, MMULTIFXPT, along=3)
            DMULTIFXPT <- DMULTIFXPT + DISFXPT
            VMULTILOGLIK <- c(VMULTILOGLIK, DNPLLOGLIK)
          }
        }
      }                           
    }       
  }

  if (dim(AMULTIFXPT)[3] > 1){
    AMULTIFXPT <- AMULTIFXPT[,,order(VMULTILOGLIK)]
    VDISTFXPT <- c()
    for (t in 1:dim(AMULTIFXPT)[3]){
      VDISTFXPT[t] <- max(abs(AMULTIFXPT[,,1]-AMULTIFXPT[,,t]))
    } 
    DDISTFXPT <- max(VDISTFXPT)
  } else{
    DDISTFXPT <- 0
  }

  if (DSUCCESSCONVTRUE==0){
    VNPLTHETA <- rep(NA, length(vTheta))
    VNPLCCPS <- rep(NA, dNumStates*dN)
    DNPLLOGLIK <- NA
    DNPLTIMEFULL <- NA
  }

  OUTSOLVENPL <- list("NPLthetasol"=VNPLTHETA, "NPLccpssol"=VNPLCCPS, "NPLloglik"=DNPLLOGLIK, "success"=DSUCCESSCONVTRUE, "numbfxpt"=DMULTIFXPT, "distfxpt"=DDISTFXPT, "timefull"=DNPLTIMEFULL)
  return(OUTSOLVENPL)
}


## Setup parallel computing, sinking and grid.
if (lParallel==TRUE){
  dCores=detectCores()
  dCl <- makeCluster(dCores[1]-1)
  registerDoParallel(dCl)  
}

if (dM==50000){
  dMC <- 1
  lJacobian <- FALSE
  lPopBR <- FALSE
} 

if (lPopBR==TRUE){
  sFileName <- paste0(sSelectDGP,"_pop")
} else{
  if (dM<1000){
    sFileName <- paste0(sSelectDGP,"_",dM)  
  } else{
    sFileName <- paste0(sSelectDGP,"_",dM/1000,"k")
  }
}

if (lSink==TRUE){
  writeLines(c(""), paste0("out_psd08_",sFileName,".txt"))
  sink(paste0("out_psd08_",sFileName,".txt"))
}

if (sSelectDGP=="grid"){
  lJacobian <- TRUE
}

if (lPopBR==TRUE) dMC <- 1


## Data generating process.

cat("Model from Pesendorfer and Schmidt-Dengler (2008, Review of Economic Studies) \n")
dN <- 2
dNumAlt <- 2
dT <- 2
if (sSelectDGP != "grid"){
  vGridThetaRN <- -2.4
  vTheta <- c(1.2,vGridThetaRN,-0.2,0.1)
} else{
  vTheta <- c(1.2,"TBA",-0.2,0.1)
  cat("\n")
  cat(paste0("Grid used for strategic interactions: from ", -min(vGridThetaRN), " to ", -max(vGridThetaRN), " by ", 0.2, " \n"))
}   
dBeta <- 0.9
vSupportX <- c()
mFS <- 1
dSizeX <- 1
dNumStates <- dSizeX*dNumAlt^dN
dFXPTol <- 10^(-6)

cat("\n")
cat("Important information: \n")
cat("\t Number of markets:", dM, "\n")
cat("\t First-stage estimation of the CCPs: ", sSelectFirstStep, "\n")
cat("\t Maximum number of iterations used for the NPL, modified NPL and spectral algorithms:", dK, "\n")
cat("\t Fixed relaxation parameter used in modified NPL algorithm:", dFixAlpha, "\n")
cat("\t Number of starting values used for iterative algorithms:", dStartVals, "\n")
cat("\t Number of Monte Carlo replications:", dMC, "\n")
cat("\t Tolerance used to assess convergence in the CCP space:", dCCPTol, "\n")
cat("\t Tolerance used to declare estimates as equivalent:", dCompTol, "\n")
cat("\t Tolerance used to assess if two fixed points are different:", dNumEqTol, "\n")

cat("\n")
cat("Information about the model: \n")
cat("\t Number of players:", dN, "\n")
cat("\t Number of periods:", dT, "\n")
cat("\t Number of alternatives:", dNumAlt, "\n")
cat("\t Number of different states:", dNumStates, "\n")
cat("\t Parameters to be estimated:", vTheta, "\n")
cat("\t Discount factor:", dBeta, "\n")

# List of all possible states.
mFullX <- matrix(data=NA, nrow=dNumStates, ncol=dN)
for (c in 1:dN){
  for (d in 0:(dNumAlt^(c-1)-1)){
    mFullX[(d*dNumAlt^(dN-c+1) + 1) : ((d+1)*dNumAlt^(dN-c+1)), c] <- cbind(rep(0,dNumAlt^(dN-c)),rep(1,dNumAlt^(dN-c)))
  }
}


# List of all possible actions.
mAN <- matrix(data = NA, nrow=dNumAlt^dN, ncol=dN)
for (h in 0:(dN-1)){
  for (l in 0:(dNumAlt^h-1)){
    mAN[(l*dNumAlt^(dN-h) + 1):((l+1)*dNumAlt^(dN-h)), h + 1] <- cbind(rep(0, dNumAlt^(dN-h-1)), rep(1, dNumAlt^(dN-h-1)))         
  }
}

# List of all opponents' possible actions.
mANm1 <- matrix(data = NA, nrow=dNumAlt^(dN-1), ncol=dN-1)
for (f in 1:(dN-1)){
  for (g in 0:(dNumAlt^(f-1)-1)){
    mANm1[(g*dNumAlt^(dN-f)+1):((g+1)*dNumAlt^(dN-f)), f] <- cbind(rep(0, dNumAlt^(dN-f-1)), rep(1, dNumAlt^(dN-f-1))) 
  }
}

# Transition matrix for all states.
mFStates <- mFS %x% matrix(rep(1,(dNumAlt^dN)^2), nrow=dNumAlt^dN, ncol=dNumAlt^dN)

dLength2step <- length(vTheta) + 1
dLengthNPLAlgo <- length(vTheta) + 11
dLengthRelaxAlgo <- length(vTheta) + 11
dLengthSpecAlgo <- length(vTheta) + 11
dLengthSpecSolver <- length(vTheta) + 4
dLengthMCMatrix <- dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + dLengthSpecSolver + (lJacobian)

aMCMatrix <- array(NA, dim=c(dLengthMCMatrix,dMC,length(vGridThetaRN)))

for (g in 1:length(vGridThetaRN)){
  if (sSelectDGP=="grid"){
    cat("\n")
    cat("Strategic interaction parameter = ", vGridThetaRN[g],"\n")
    vTheta <- c(1.2,-vGridThetaRN[g],-0.2,0.1)
  } 

  # Equilibrium CCPs in DGP.
  outSolveDGPCCPs <- fMultiSolveDGPCCPs(dSEqDGP)
  aTrueCCPs <- outSolveDGPCCPs$trueccps
  dSuccessConvTrue <- outSolveDGPCCPs$success
  if (dim(aTrueCCPs)[3]>1){
    cat("\n")
    cat("The solver found", dim(aTrueCCPs)[3], "equilibria among ", dSEqDGP, " randomly picked starting values.\n")
    aSortedTrueCCPs <- aTrueCCPs[,,order(aTrueCCPs[1,1,], decreasing=TRUE)]
    cat("Largest distance between chosen equilibrium and other equilibria as mesured by the largest supremum norm = ", outSolveDGPCCPs$disteq)
  } else{
    cat("\n")
    cat("The solver found a single equilibrium among", dSEqDGP, "randomly picked starting values. \n")
  }
  cat("\n")
  cat("Failed convergence when solving for true model =", dSEqDGP-dSuccessConvTrue,"out of", dSEqDGP,". \n")
  cat("\n")
  cat("True CCPs (one column for each player) = \n");
  if (sSelectDGP=="i"){
    mTrueCCPs <- aSortedTrueCCPs[,,1]
  } else if (sSelectDGP=="ii"){
    mTrueCCPs <- aSortedTrueCCPs[,,2]
  } else if (sSelectDGP=="iii"){
    mTrueCCPs <- aSortedTrueCCPs[,,3]
  }
  print(mTrueCCPs)

  # Compute the steady state distribution of the state variables.
  mFPANtrue <- fFP(mTrueCCPs)
  vSSX0 <- (1/dNumStates)*rep(1,dNumStates)
  dDist <- 1
  while (dDist > dFXPTol){
    vSSX <- t(mFPANtrue)%*%vSSX0
    dDist <- max(abs(vSSX - vSSX0))
    vSSX0 = vSSX  
  }


  ## Monte Carlo.

  cat("\n");
  cat("Progress of the Monte Carlo: \n")
  if (lSink==TRUE) sink()

  mMCMatrix <-foreach(mc=1:dMC, .combine=cbind, .packages=c('MASS','matrixStats','BB','abind','numDeriv')) %dopar% {  
    if (lSink==TRUE) sink(paste0("out_psd08_",sFileName,".txt"), append=TRUE)
    cat(paste("\t ", (mc/dMC)*100, "% \n"))
    set.seed(1234+mc)
    
    # Given the steady state distribution and the true CCPs, draw a sample of markets.
    mInitial <- matrix(runif((dN+1)*dM, min = 0, max = 1), nrow=dN+1, ncol=dM)
    vCumul <- cumsum(cbind(0,t(vSSX)))  
    mA0 <- matrix(data=NA, nrow=dM, ncol=dN)
    mA1 <- matrix(data=NA, nrow=dM, ncol=dN)
    vIndState <- matrix(data=NA, nrow=dM, ncol=1)   
    for (m in 1:dM){
      vSelect <- (mInitial[1,m]*rep(1,dNumStates+1)>=vCumul)  *  (mInitial[1,m]*rep(1,dNumStates+1)<t(cbind(t(vCumul[2:(dNumStates+1)]), 1)))  *  (1:(dNumStates+1)) 
      mA0[m,] = mFullX[max(vSelect),1:dN]
      mA1[m,] = (t(mInitial[2:(dN+1),m])<=mTrueCCPs[max(vSelect),])
      vIndState[m] = max(vSelect)
    }
    mX <- cbind(mA0)

    # Estimation of the first-stage CCPs.
    if (sSelectFirstStep=="true"){
      mCCPhat <- mTrueCCPs
    } else if (sSelectFirstStep=="kernel"){
      vH <- (4/(dim(mX)[2]+2))^(1/(dim(mX)[2]+4))*sqrt(colVars(mX))*(dM^(-1/(dim(mX)[2]+4)))
      aDiffX <- array(data=NA, dim=c(dM,dNumStates,dim(mX)[2]))
      for (x in 1:dim(mX)[2]){
        aDiffX[,,x] <- (1/vH[x]^2)*(matrix(mX[,x], dM, dNumStates)-matrix(mFullX[,x], dM, dNumStates, byrow=TRUE))^2
      }
      mK <- (2*pi)^(-0.5*dim(mX)[2])*exp(-0.5*apply(aDiffX, MARGIN=c(1,2), sum))
      mKW <- mK%*%diag(1/colSums(mK))
      mCCPhat <- t(t(mA1)%*%mKW)
    } else if (sSelectFirstStep=="freq"){ 
      aIndX <- array(data=NA, dim=c(dM,dNumStates,dim(mX)[2]))
      for (x in 1:dim(mX)[2]){
        aIndX[,,x] <- (matrix(mX[,x], dM, dNumStates)==matrix(mFullX[,x], dM, dNumStates, byrow=TRUE))
      }
      mK <- apply(aIndX, MARGIN=c(1,2), prod)
      mKW <- mK%*%diag(1/colMaxs(rbind(colSums(mK),rep(1,dNumStates))))
      mCCPhat <- t(t(mA1)%*%mKW)
    }else {
      cat("\n")
      cat("WARNING: NO FIRST-STAGE METHOD SELECTED.\n")
    }
    mCCPhat <- (mCCPhat>0)*(mCCPhat<1)*mCCPhat + (mCCPhat>=1)*(1-10^(-10)) + (mCCPhat<=0)*10^(-10)

    # Competition summary statistics.
    if (dMC == 1 & dM == 50000){
      cat("\n")
      cat("Competition statistics.\n")
      cat("\t Average number of active firms = ",mean(rowSums(mA1)),"\n")
      cat("\t Std. Dev. number of active firms = ",sqrt(var(rowSums(mA1))),"\n")

      mARReg <- cbind(rep(1,dM), rowSums(mA0))
      vARPar <- solve(t(mARReg)%*%mARReg)%*%t(mARReg)%*%rowSums(mA1)
      cat("\t AR(1) parameter for number of active firms = ", vARPar[2],"\n")      

      vNumbEntrants <- rowSums((mA1 - mA0)>0)
      vNumbExits <- rowSums((mA1 - mA0)<0)
      cat("\t Average number of entrants = ", mean(vNumbEntrants),"\n")
      cat("\t Average number of exits = ", mean(vNumbExits),"\n")      
      cat("\t Average excess turnover = ", mean(vNumbEntrants + vNumbExits - abs(vNumbEntrants - vNumbExits)),"\n")
      cat("\t Correlation between entries and exits = ", cor(vNumbEntrants, vNumbExits),"\n")

      cat("\t Probability of being active = ",colMeans(mA1),"\n")
    }

    # 2-step estimation.
    if (sSelectDGP!="grid" & dM != 50000){
      out2step <- f2step(mCCPhat)
      vTheta2step <- out2step$thetaest
      dLogLik2step <- out2step$loglikest
    } else{
      vTheta2step <- rep(NA, length(vTheta))
      dLogLik2step <- NA
    }
    vStore2step <- c(vTheta2step, dLogLik2step)

    # NPL algorithm.
    if (sSelectDGP!="grid" & dM != 50000){
      outMultiNPLAlgo <- fMultiNPLAlgo(dK, dStartVals, mCCPhat)
      vThetaNPLLast <- outMultiNPLAlgo$thetalast
      vCCPNPLLast <- outMultiNPLAlgo$ccplast
      dLogLikNPLLast <- outMultiNPLAlgo$logliklast
      dDiffCCPsNPLLast <- outMultiNPLAlgo$diffccpslast
      dKNPLLast <- outMultiNPLAlgo$k
      dNumbFxPtNPLAlgo <- outMultiNPLAlgo$numbfxpt   
      dDistFxPtNPLAlgo <- outMultiNPLAlgo$distfxpt
      vAvgTimeNPL <- outMultiNPLAlgo$avgtimes
      vSdTimeNPL <- outMultiNPLAlgo$sdtimes
    } else{
      vThetaNPLLast <- rep(NA, length(vTheta))
      vCCPNPLLast <- rep(NA, length(mTrueCCPs))
      dLogLikNPLLast <- NA
      dDiffCCPsNPLLast <- NA
      dKNPLLast <- NA
      dNumbFxPtNPLAlgo <- NA
      dDistFxPtNPLAlgo <- NA
      vAvgTimeNPL <- rep(NA, 3)
      vSdTimeNPL <- rep(NA, 3)
    }
    vStoreNPLAlgo <- c(vThetaNPLLast, dLogLikNPLLast, dDiffCCPsNPLLast, dKNPLLast, dNumbFxPtNPLAlgo, dDistFxPtNPLAlgo, vAvgTimeNPL, vSdTimeNPL)

    # Relaxation algorithm.
    if (sSelectDGP!="grid" & dM != 50000){
      outMultiRelaxAlgo <- fMultiRelaxAlgo(dK, dStartVals, mCCPhat, dFixAlpha)
      vThetaRelaxLast <- outMultiRelaxAlgo$thetalast
      vCCPRelaxLast <- outMultiRelaxAlgo$ccplast
      dLogLikRelaxLast <- outMultiRelaxAlgo$logliklast 
      dDiffCCPsRelaxLast <- outMultiRelaxAlgo$diffccpslast
      dKRelaxLast <- outMultiRelaxAlgo$k
      dNumbFxPtRelaxAlgo <- outMultiRelaxAlgo$numbfxpt        
      dDistFxPtRelaxAlgo <- outMultiRelaxAlgo$distfxpt 
      vAvgTimeRelax <- outMultiRelaxAlgo$avgtimes
      vSdTimeRelax <- outMultiRelaxAlgo$sdtimes             
    } else{
      vThetaRelaxLast <- rep(NA, length(vTheta))
      vCCPRelaxLast <- rep(NA, length(mTrueCCPs))
      dLogLikRelaxLast <- NA
      dDiffCCPsRelaxLast <- NA
      dKRelaxLast <- NA
      dNumbFxPtRelaxAlgo <- NA
      dDistFxPtRelaxAlgo <- NA
      vAvgTimeRelax <- rep(NA, 3)
      vSdTimeRelax <- rep(NA, 3)      
    }
    vStoreRelaxAlgo <- c(vThetaRelaxLast, dLogLikRelaxLast, dDiffCCPsRelaxLast, dKRelaxLast, dNumbFxPtRelaxAlgo, dDistFxPtRelaxAlgo, vAvgTimeRelax, vSdTimeRelax)

    # Spectral algorithm.
    if (sSelectDGP!="grid" & dM != 50000){
      outMultiSpecAlgo <- fMultiSpecAlgo(dK, dStartVals, mCCPhat)
      vThetaSpecLast <- outMultiSpecAlgo$thetalast
      vCCPSpecLast <- outMultiSpecAlgo$ccplast
      dLogLikSpecLast <- outMultiSpecAlgo$logliklast
      dDiffCCPsSpecLast <- outMultiSpecAlgo$diffccpslast
      dKSpecLast <- outMultiSpecAlgo$k
      dNumbFxPtSpecAlgo <- outMultiSpecAlgo$numbfxpt   
      dDistFxPtSpecAlgo <- outMultiSpecAlgo$distfxpt
      vAvgTimeSpec <- outMultiSpecAlgo$avgtimes
      vSdTimeSpec <- outMultiSpecAlgo$sdtimes
    } else{
      vThetaSpecLast <- rep(NA, length(vTheta))
      vCCPSpecLast <- rep(NA, length(mTrueCCPs))
      dLogLikSpecLast <- NA
      dDiffCCPsSpecLast <- NA
      dKSpecLast <- NA
      dNumbFxPtSpecAlgo <- NA
      dDistFxPtSpecAlgo <- NA
      vAvgTimeSpec <- rep(NA, 3)
      vSdTimeSpec <- rep(NA, 3)
    }
    vStoreSpecAlgo <- c(vThetaSpecLast, dLogLikSpecLast, dDiffCCPsSpecLast, dKSpecLast, dNumbFxPtSpecAlgo, dDistFxPtSpecAlgo, vAvgTimeSpec, vSdTimeSpec)

    # Spectral solver.
    if (dM != 50000){
      outMultiSolveNPL <- fMultiSolveNPL(dStartVals, mCCPhat)
      vThetaSolveNPL <- outMultiSolveNPL$NPLthetasol
      vSolveNPLCCPs <- outMultiSolveNPL$NPLccpssol
      dLogLikSolveNPL <- outMultiSolveNPL$NPLloglik
      dNumbFxPtSolveNPL <- outMultiSolveNPL$numbfxpt 
      dDistFxPtSolveNPL <- outMultiSolveNPL$distfxpt
      dTimeSolveNPL <- outMultiSolveNPL$timefull       
      if (outMultiSolveNPL$success==0){
        cat("\n")
        cat("WARNING: SPECTRAL SOLVER NEVER CONVERGED AT ", mc,".\n")      
      }       
    } else{
      vThetaSolveNPL <- rep(NA, length(vTheta))
      vSolveNPLCCPs <- rep(NA, length(mTrueCCPs))
      dLogLikSolveNPL <- NA
      dNumbFxPtSolveNPL <- NA
      dDistFxPtSolveNPL <- NA
      dTimeSolveNPL <- NA      
    }
    vStoreSpecSolver <- c(vThetaSolveNPL, dLogLikSolveNPL, dNumbFxPtSolveNPL, dDistFxPtSolveNPL, dTimeSolveNPL)
    
    # Jacobian at spectral solver estimates.
    if (lJacobian){
      dIdxBestFXLogLik <- NA
      vConvInd <- c(dKNPLLast<dK, dKRelaxLast<dK, dKSpecLast<dK, outMultiSolveNPL$success!=0)
      if (sum(vConvInd)>0){
        mRestrTheta <- cbind(vThetaNPLLast, vThetaRelaxLast, vThetaSpecLast, vThetaSolveNPL)
        mRestrCCP <- cbind(vCCPNPLLast, vCCPRelaxLast, vCCPSpecLast, vSolveNPLCCPs)
        vLogLik <- c(dLogLikNPLLast, dLogLikRelaxLast, dLogLikSpecLast, dLogLikSolveNPL)
        vConvLogLik <- vLogLik*vConvInd
        vConvLogLik[vConvLogLik==0] <- NA
        dBestFXLogLik <- min(vConvLogLik, na.rm=TRUE)
        dIdxBestFXLogLik <- which.min(vConvLogLik)
        vBestTheta <- mRestrTheta[,dIdxBestFXLogLik]
        vBestCCP <- mRestrCCP[,dIdxBestFXLogLik]
      }
      if (!is.na(dIdxBestFXLogLik)){
        mJacobian <- jacobian(fJacobian, vSolveNPLCCPs, method="simple", side=NULL, method.args=list())
        dSpecRadNPL <- max(abs(eigen(mJacobian)$values))
      } else{
        dSpecRadNPL <- NA
      }      
    } else{
      dSpecRadNPL <- c()
    }

    if (lSink==TRUE) sink()
    
    # Store results
    invisible(c(vStore2step, vStoreNPLAlgo, vStoreRelaxAlgo, vStoreSpecAlgo, vStoreSpecSolver, dSpecRadNPL))
  }

  mMCMatrix <- as.matrix(mMCMatrix)
  aMCMatrix[,,g] <- mMCMatrix

  if (lSink==TRUE) sink(paste0("out_psd08_",sFileName,".txt"), append = TRUE)
  if (lSaveMCMatrix==TRUE) save(aMCMatrix, file=paste0("matrix_",sFileName,".rda"))    

  ## Rearrange MC matrix.
  mMCMatrix <- as.matrix(aMCMatrix[,,g])

  mTheta2stepMC <- mMCMatrix[1:length(vTheta),]
  vLogLik2stepMC <- mMCMatrix[length(vTheta)+1,]

  mThetaNPLLastMC <- mMCMatrix[(dLength2step + 1):(dLength2step + length(vTheta)),]
  vLogLikNPLLastMC <- mMCMatrix[(dLength2step + length(vTheta)+1),]
  vDiffCCPsNPLLastMC <- mMCMatrix[(dLength2step + length(vTheta)+2),]
  vKNPLLastMC <- mMCMatrix[(dLength2step + length(vTheta)+3),]
  vNumbFxPtNPLAlgoMC <- mMCMatrix[(dLength2step + length(vTheta)+4),]
  vDistFxPtNPLAlgoMC <- mMCMatrix[(dLength2step + length(vTheta)+5),]
  mAvgTimeNPLMC <- mMCMatrix[(dLength2step + length(vTheta)+6):(dLength2step + length(vTheta)+8),]
  vTimeNPLMC <- mAvgTimeNPLMC[1,]*vKNPLLastMC
  vTimeNPLMC[vKNPLLastMC==dK] <- NA
  mAvgTimeNPLMC <- rbind(mAvgTimeNPLMC, vTimeNPLMC)

  mThetaRelaxLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + 1):(dLength2step + dLengthNPLAlgo + length(vTheta)),]
  vLogLikRelaxLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+1),]
  vDiffCCPsRelaxLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+2),]
  vKRelaxLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+3),]
  vNumbFxPtRelaxAlgoMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+4),]
  vDistFxPtRelaxAlgoMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+5),]
  mAvgTimeRelaxMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + length(vTheta)+6):(dLength2step + dLengthNPLAlgo + length(vTheta)+8),]
  vTimeRelaxMC <- mAvgTimeRelaxMC[1,]*vKRelaxLastMC
  vTimeRelaxMC[vKRelaxLastMC==dK] <- NA
  mAvgTimeRelaxMC <- rbind(mAvgTimeRelaxMC, vTimeRelaxMC)

  mThetaSpecLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + 1):(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)),]
  vLogLikSpecLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+1),]
  vDiffCCPsSpecLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+2),]
  vKSpecLastMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+3),]
  vNumbFxPtSpecAlgoMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+4),]
  vDistFxPtSpecAlgoMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+5),]
  mAvgTimeSpecMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+6):(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + length(vTheta)+8),]
  vTimeSpecMC <- mAvgTimeSpecMC[1,]*vKSpecLastMC
  vTimeSpecMC[vKSpecLastMC==dK] <- NA
  mAvgTimeSpecMC <- rbind(mAvgTimeSpecMC, vTimeSpecMC)

  mThetaSolveNPLMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + 1):(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + length(vTheta)),]
  vLogLikSolveNPLMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + length(vTheta) + 1),]
  vNumbFxPtSolveNPLMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + length(vTheta) + 2),]
  vDistFxPtSolveNPLMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + length(vTheta) + 3),]
  vTimeSolveNPLMC <- mMCMatrix[(dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + length(vTheta) + 4),]
  vNotNASolveNPLMC <- !is.na(vLogLikSolveNPLMC)
  dNotNASolveNPLMC <- sum(vNotNASolveNPLMC)

  if (lJacobian) vSpecRadNPL <- mMCMatrix[dLength2step + dLengthNPLAlgo + dLengthRelaxAlgo + dLengthSpecAlgo + dLengthSpecSolver + 1, ]

  ## Summarizing results.

  if (dM != 50000){
    if (sSelectDGP!="grid"){
      
      # Convergence of K-step NPL algorithm.
      if (dMC>1){
        vKNPLEst <- rowMeans(mThetaNPLLastMC)
        vKNPLStdErr <- rowSds(mThetaNPLLastMC)
        vConvKNPL <- vKNPLLastMC<dK
      } else{
        vKNPLEst <- as.matrix(mThetaNPLLastMC)
        vKNPLStdErr <- matrix(0, nrow=nrow(vKNPLEst), ncol=ncol(vKNPLEst))
        vConvKNPL <- vKNPLLastMC<dK
      }
      dConvKNPL <- sum(vConvKNPL, na.rm=TRUE)
      cat("\n")
      cat(paste("Ratio of converged sequences for the NPL algorithm (max absolute distance in CCP space being less than ", dCCPTol,"before K = ", dK,"):", dConvKNPL/dMC*100,"% of Monte Carlo samples. \n"))

      # Convergence of K-step relaxation algorithm.
      if (dMC>1){
        vKRelaxEst <- rowMeans(mThetaRelaxLastMC)
        vKRelaxStdErr <- rowSds(mThetaRelaxLastMC)
        vConvKRelax <- vKRelaxLastMC<dK
      } else{
        vKRelaxEst <- as.matrix(mThetaRelaxLastMC)
        vKRelaxStdErr <- matrix(0, nrow=nrow(vKRelaxEst), ncol=ncol(vKRelaxEst))
        vConvKRelax <- vKRelaxLastMC<dK
      }
      dConvKRelax <- sum(vConvKRelax, na.rm=TRUE)
      cat("\n")
      cat(paste("Ratio of converged sequences for the relaxation algorithm (max absolute distance in CCP space being less than ", dCCPTol,"before K = ", dK,"):", dConvKRelax/dMC*100,"% of Monte Carlo samples. \n"))

      # Convergence of K-step spectral algorithm.
      if (dMC>1){
        vKSpecEst <- rowMeans(mThetaSpecLastMC)
        vKSpecStdErr <- rowSds(mThetaSpecLastMC)
        vConvKSpec <- vKSpecLastMC<dK
      } else{
        vKSpecEst <- as.matrix(mThetaSpecLastMC)
        vKSpecStdErr <- matrix(0, nrow=nrow(vKSpecEst), ncol=ncol(vKSpecEst))
        vConvKSpec <- vKSpecLastMC<dK
      }
      dConvKSpec <- sum(vConvKSpec, na.rm=TRUE)
      cat("\n")
      cat(paste("Ratio of converged sequences for the spectral algorithm (max absolute distance in CCP space being less than ", dCCPTol,"before K = ", dK,"):", dConvKSpec/dMC*100,"% of Monte Carlo samples. \n"))

      # Convergence and spectral radius.
      if (lJacobian==TRUE){
        cat("\n")
        cat("Ratio of samples where spectral solver did not converge = ", (dMC-dNotNASolveNPLMC)/dMC*100, "% \n")
        
        cat("\n")
        cat("Ratio of samples where spectral radius is smaller than 1 at converged estimates maximizing the likelihood = ", sum((vSpecRadNPL<1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t NPL algorithm converged = ", sum((vConvKNPL)*(vSpecRadNPL<1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Relaxation algorithm converged = ", sum((vConvKRelax)*(vSpecRadNPL<1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Spectral algorithm converged = ", sum((vConvKSpec)*(vSpecRadNPL<1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Spectral solver converged = ", sum((vNotNASolveNPLMC)*(vSpecRadNPL<1),na.rm=TRUE)/dMC*100, "% \n")

        cat("\n")
        cat("Ratio of samples where spectral radius is greater than 1 at converged estimates maximizing the likelihood = ", sum((vSpecRadNPL>=1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t NPL algorithm converged = ", sum((vConvKNPL)*(vSpecRadNPL>=1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Relaxation algorithm converged = ", sum((vConvKRelax)*(vSpecRadNPL>=1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Spectral algorithm converged = ", sum((vConvKSpec)*(vSpecRadNPL>=1),na.rm=TRUE)/dMC*100, "% \n")
        cat("\t Spectral solver converged = ", sum((vNotNASolveNPLMC)*(vSpecRadNPL>=1),na.rm=TRUE)/dMC*100, "% \n")

        if (sum((1-vConvKNPL)*(vSpecRadNPL<1),na.rm=TRUE)>0){
          vIndKNPLNotConvStable <- ((1-vConvKNPL)*(vSpecRadNPL<1))==1
          vIndKNPLNotConvStable[is.na(vIndKNPLNotConvStable)] <- FALSE
          vDiffCCPsNPLNotConvStable <- vDiffCCPsNPLLastMC[vIndKNPLNotConvStable]
          cat("\n")
          cat("Distance in CCP space when K-step NPL algorithm does not converge and spectral radius smaller than 1: \n")
          cat("\t Minimum = ", min(vDiffCCPsNPLNotConvStable),"\n")
          cat("\t Median = ", median(vDiffCCPsNPLNotConvStable),"\n")
          cat("\t Maximum = ", max(vDiffCCPsNPLNotConvStable),"\n")   
          cat("\n")
          cat("Average value of spectral radius when K-step NPL algorithm does not converge and spectral radius smaller than 1: ",mean(vSpecRadNPL[vIndKNPLNotConvStable])," \n")         
        }
      }

      # Assessing best fixed points.
      mLogLikMC <- cbind(vLogLikNPLLastMC*vConvKNPL,vLogLikRelaxLastMC*vConvKRelax,vLogLikSpecLastMC*vConvKSpec,vLogLikSolveNPLMC*vNotNASolveNPLMC)
      mLogLikMC[mLogLikMC==0] <- NA

      vWhichBestFXLogLik <- apply(mLogLikMC, MARGIN=1, which.min)

      aThetaAllMC <- abind(t(mThetaNPLLastMC)*vConvKNPL, t(mThetaRelaxLastMC)*vConvKRelax, t(mThetaSpecLastMC)*vConvKSpec, t(mThetaSolveNPLMC)*vNotNASolveNPLMC, along=3)
      aThetaAllMC[aThetaAllMC==0] <- NA
      mDiffBestThetaAllMC <- matrix(NA, nrow=dim(aThetaAllMC)[1], ncol=dim(aThetaAllMC)[3])
      for (mc in 1:dim(mMCMatrix)[2]){
        vBestEst <- aThetaAllMC[mc, ,vWhichBestFXLogLik[mc]]
        for (a in 1:dim(aThetaAllMC)[3]){
          mDiffBestThetaAllMC[mc,a] <- max(abs(aThetaAllMC[mc,,a]-vBestEst))<dCompTol
        }
      }

      cat("\n")
      cat("Ratio of samples in which each estimator reaches the estimates (up to maximum absolute difference being less than ", dCompTol,") that maximize log-likelihood. \n")
      cat("\t K =", dK, "NPL algorithm = ",sum(mDiffBestThetaAllMC[,1], na.rm=TRUE)/dMC*100,"% \n")
      cat("\t K =", dK, "relaxation algorithm = ",sum(mDiffBestThetaAllMC[,2], na.rm=TRUE)/dMC*100,"% \n")
      cat("\t K =", dK, "spectral algorithm = ",sum(mDiffBestThetaAllMC[,3], na.rm=TRUE)/dMC*100,"% \n")
      cat("\t Spectral solver = ",sum(mDiffBestThetaAllMC[,4], na.rm=TRUE)/dMC*100,"% \n")

      # Properties of 2-step estimates (all samples).  
      vResultsColNames <- c("True", "Avg", "Std errors")
      if (dMC>1){
        v2stepEst <- rowMeans(mTheta2stepMC)
        v2stepStdErr <- rowSds(mTheta2stepMC)
      } else{
        v2stepEst <- as.matrix(mTheta2stepMC)
        v2stepStdErr <- matrix(0, nrow=nrow(vKNPLEst), ncol=ncol(vKNPLEst))
      }      
      mResults2step <- cbind(vTheta, v2stepEst, v2stepStdErr)
      colnames(mResults2step)<-vResultsColNames
      cat("\n")
      cat("Average 2-step estimates and standard errors from all samples \n")
      print(as.matrix(mResults2step))

      # Properties of K-step NPL algorithm estimates (all samples).  
      mResultsAllKNPL <- cbind(vTheta, vKNPLEst, vKNPLStdErr)
      colnames(mResultsAllKNPL)<-vResultsColNames
      cat("\n")
      cat("Average NPL algorithm estimates and standard errors from all samples at K = ", dK," \n")
      print(as.matrix(mResultsAllKNPL))

      # Properties of K-step relaxation algorithm estimates (all samples).  
      mResultsAllKRelax <- cbind(vTheta, vKRelaxEst, vKRelaxStdErr)
      colnames(mResultsAllKRelax)<-vResultsColNames
      cat("\n")
      cat("Average relaxation algorithm estimates and standard errors from all samples at K = ", dK," \n")
      print(as.matrix(mResultsAllKRelax))

      # Properties of K-step spectral algorithm estimates (all samples).  
      mResultsAllKSpec <- cbind(vTheta, vKSpecEst, vKSpecStdErr)
      colnames(mResultsAllKSpec)<-vResultsColNames
      cat("\n")
      cat("Average spectral algorithm estimates and standard errors from all samples at K = ", dK," \n")
      print(as.matrix(mResultsAllKSpec))

      # Properties of K-step NPL algorithm estimates (converged samples).  
      if (dConvKNPL>1 & dMC>1){
        vKNPLEstConv <- rowMeans(mThetaNPLLastMC[,vConvKNPL])
        vKNPLStdErrConv <- rowSds(mThetaNPLLastMC[,vConvKNPL])
        mResultsConvKNPL <- cbind(vTheta, vKNPLEstConv, vKNPLStdErrConv)
        colnames(mResultsConvKNPL) <- vResultsColNames
        cat("\n")
        cat("Average NPL algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        print(as.matrix(mResultsConvKNPL))
      } else if (dMC>1){
        cat("\n")
        cat("Average NPL algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        cat("NA: At most one sequence converged. \n")
      }

      # Properties of K-step relaxation algorithm estimates (converged samples).  
      if (dConvKRelax>1 & dMC>1){
        vKRelaxEstConv <- rowMeans(mThetaRelaxLastMC[,vConvKRelax])
        vKRelaxStdErrConv <- rowSds(mThetaRelaxLastMC[,vConvKRelax])
        mResultsConvKRelax <- cbind(vTheta, vKRelaxEstConv, vKRelaxStdErrConv)
        colnames(mResultsConvKRelax) <- vResultsColNames
        cat("\n")
        cat("Average relaxation algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        print(as.matrix(mResultsConvKRelax))
      } else if (dMC>1){
        cat("\n")
        cat("Average relaxation algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        cat("NA: At most one sequence converged. \n")
      }
    
      # Properties of K-step spectral algorithm estimates (converged samples).  
      if (dConvKSpec>1 & dMC>1){
        vKSpecEstConv <- rowMeans(mThetaSpecLastMC[,vConvKSpec])
        vKSpecStdErrConv <- rowSds(mThetaSpecLastMC[,vConvKSpec])
        mResultsConvKSpec <- cbind(vTheta, vKSpecEstConv, vKSpecStdErrConv)
        colnames(mResultsConvKSpec) <- vResultsColNames
        cat("\n")
        cat("Average spectral algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        print(as.matrix(mResultsConvKSpec))
      } else if (dMC>1){
        cat("\n")
        cat("Average spectral algorithm estimates and standard errors from converged samples at K = ", dK," \n")
        cat("NA: At most one sequence converged. \n")
      }

      if (lJacobian==TRUE){
        # Properties of the conditional estimator.
        if (sum(vSpecRadNPL<1, na.rm=TRUE)>1 & dMC>1){
          vKNPLEstStable <- rowMeans(mThetaNPLLastMC[,vSpecRadNPL<1], na.rm=TRUE)
          vKNPLStdErrStable <- rowSds(mThetaNPLLastMC[,vSpecRadNPL<1], na.rm=TRUE)
          mResultsStableKNPL <- cbind(vTheta, vKNPLEstStable, vKNPLStdErrStable)
          colnames(mResultsStableKNPL) <- vResultsColNames
          cat("\n")
          cat("Average NPL algorithm estimates and standard errors from stable samples at K = ", dK," \n")
          print(as.matrix(mResultsStableKNPL))
        } else if (dMC>1){
          cat("\n")
          cat("Average NPL algorithm estimates and standard errors from stable samples at K = ", dK," \n")
          cat("NA: At most one stable samples. \n")
        }
      }
    }

    # Properties of the solver algorithm estimates.
    vResultsColNames <- c("True", "Avg", "Std errors")
    cat("\n")
    cat("Ratio of Monte-Carlo samples for which the solver converged =", dNotNASolveNPLMC/dMC*100, "%. \n")
    if (dMC>1){
      mResultsSolveNPL <- cbind(vTheta, rowMeans(mThetaSolveNPLMC, na.rm=TRUE), rowSds(mThetaSolveNPLMC, na.rm=TRUE))    
    } else{
      mResultsSolveNPL <- cbind(vTheta, mThetaSolveNPLMC, rep(0, length(vTheta)))
    }
    cat("\n")
    cat("Average solver algorithm estimates and standard errors = \n")
    colnames(mResultsSolveNPL) <- vResultsColNames
    print(as.matrix(mResultsSolveNPL))

    if (sSelectDGP!="grid"){
      
      if (dNotNASolveNPLMC>1){
        if (dMC>1){
          vIndMatchSpecSolve <- colMaxs(abs(mThetaSpecLastMC[,vNotNASolveNPLMC] - mThetaSolveNPLMC[,vNotNASolveNPLMC]))<dCompTol    
        } else {
          vIndMatchSpecSolve <- max(abs(mThetaSpecLastMC[,vNotNASolveNPLMC] - mThetaSolveNPLMC[,vNotNASolveNPLMC]))<dCompTol
        } 
        dMatchSpecSolve <- sum(vIndMatchSpecSolve, na.rm=TRUE)
        cat("\n")
        cat("Ratio of samples where K-step spectral algorithm estimates coincide with converged solver algorithm NPL estimates (max distance being less than ", dCompTol ,") = ", dMatchSpecSolve/dNotNASolveNPLMC*100, "% \n")
      }

      # Average number of iterations to converge.
      cat("\n")
      cat("Number of iterations to converge (conditional on convergence): \n")
      cat("\t NPL algorithm: mean = ", mean(vKNPLLastMC[vKNPLLastMC<dK]), ", standard deviation = ", sd(vKNPLLastMC[vKNPLLastMC<dK]), " . \n")
      cat("\t Relaxation algorithm: mean = ", mean(vKRelaxLastMC[vKRelaxLastMC<dK]), ", standard deviation = ", sd(vKRelaxLastMC[vKRelaxLastMC<dK]), " . \n")
      cat("\t Spectral algorithm: mean = ", mean(vKSpecLastMC[vKSpecLastMC<dK]), ", standard deviation = ", sd(vKSpecLastMC[vKSpecLastMC<dK]), " . \n")

      # Average computation time.
      vTimeRowNames <- c("Full iteration", "Update theta", "Update CCP", "Total")
      vTimeColNames <- c("Avg", "Std dev")

      mTimeNPLMC <- cbind(rowMeans(mAvgTimeNPLMC, na.rm=TRUE), rowSds(mAvgTimeNPLMC, na.rm=TRUE))
      rownames(mTimeNPLMC) <- vTimeRowNames
      colnames(mTimeNPLMC) <- vTimeColNames
      cat("\n")
      cat("Computation time in seconds for NPL algorithm (across Monte Carlo samples): \n") 
      print(as.matrix(round(mTimeNPLMC,6)))

      mTimeRelaxMC <- cbind(rowMeans(mAvgTimeRelaxMC, na.rm=TRUE), rowSds(mAvgTimeRelaxMC, na.rm=TRUE))
      rownames(mTimeRelaxMC) <- vTimeRowNames
      colnames(mTimeRelaxMC) <- vTimeColNames
      cat("\n")
      cat("Computation time in seconds for relaxation algorithm (averaged across Monte Carlo samples): \n") 
      print(as.matrix(round(mTimeRelaxMC,6)))

      mTimeSpecMC <- cbind(rowMeans(mAvgTimeSpecMC, na.rm=TRUE), rowSds(mAvgTimeSpecMC, na.rm=TRUE))
      rownames(mTimeSpecMC) <- vTimeRowNames
      colnames(mTimeSpecMC) <- vTimeColNames
      cat("\n")
      cat("Computation time in seconds for spectral algorithm (averaged across Monte Carlo samples): \n") 
      print(as.matrix(round(mTimeSpecMC,6)))

      cat("\n")
      cat("Computation time in seconds for spectral solver (upon convergence): avg = ", mean(vTimeSolveNPLMC[vNotNASolveNPLMC]), ", std dev = ", sd(vTimeSolveNPLMC[vNotNASolveNPLMC]), " \n")
      cat("\n")

      # Instances and maximum distances of multiple fixed points in samples.
      vStatsNames <- c("Min", "Median", "Max")
  
      cat("\n")
      cat("Checking for multiple fixed points in samples via NPL algorithm: \n")
      print.table(table(vNumbFxPtNPLAlgoMC, dnn="Number of fixed points found"))
      if (max(vNumbFxPtNPLAlgoMC)>1){
        mStatsDistFxPtNPLAlgoMC <- cbind(min(vDistFxPtNPLAlgoMC[vNumbFxPtNPLAlgoMC>1]), median(vDistFxPtNPLAlgoMC[vNumbFxPtNPLAlgoMC>1]), max(vDistFxPtNPLAlgoMC[vNumbFxPtNPLAlgoMC>1]))
        colnames(mStatsDistFxPtNPLAlgoMC) <- vStatsNames
        cat("Summary statistics of furthest equilibria:\n")
        print(as.matrix(mStatsDistFxPtNPLAlgoMC))        
      }

      cat("\n")
      cat("Checking for multiple fixed points in samples via relaxation algorithm:\n")
      print.table(table(vNumbFxPtRelaxAlgoMC, dnn="Number of fixed points found"))
      if (max(vNumbFxPtRelaxAlgoMC)>1){
        mStatsDistFxPtRelaxAlgoMC <- cbind(min(vDistFxPtRelaxAlgoMC[vNumbFxPtRelaxAlgoMC>1]), median(vDistFxPtRelaxAlgoMC[vNumbFxPtRelaxAlgoMC>1]), max(vDistFxPtRelaxAlgoMC[vNumbFxPtRelaxAlgoMC>1]))
        colnames(mStatsDistFxPtRelaxAlgoMC) <- vStatsNames
        cat("Summary statistics of furthest equilibria:\n")
        print(as.matrix(mStatsDistFxPtRelaxAlgoMC))        
      }        

      cat("\n")
      cat("Checking for multiple fixed points in samples via spectral algorithm:\n")
      print.table(table(vNumbFxPtSpecAlgoMC, dnn="Number of fixed points found"))
      if (max(vNumbFxPtSpecAlgoMC)>1){
        mStatsDistFxPtSpecAlgoMC <- cbind(min(vDistFxPtSpecAlgoMC[vNumbFxPtSpecAlgoMC>1]), median(vDistFxPtSpecAlgoMC[vNumbFxPtSpecAlgoMC>1]), max(vDistFxPtSpecAlgoMC[vNumbFxPtSpecAlgoMC>1]))
        colnames(mStatsDistFxPtSpecAlgoMC) <- vStatsNames
        cat("Summary statistics of furthest equilibria:\n")
        print(as.matrix(mStatsDistFxPtSpecAlgoMC))        
      }

      cat("\n")
      cat("Checking for multiple fixed points in samples via spectral solver:\n")
      print.table(table(vNumbFxPtSolveNPLMC, dnn="Number of fixed points found"))
      if (max(vNumbFxPtSolveNPLMC)>1){
        mStatsDistFxPtSolveNPLMC <- cbind(min(vDistFxPtSolveNPLMC[vNumbFxPtSolveNPLMC>1]), median(vDistFxPtSolveNPLMC[vNumbFxPtSolveNPLMC>1]), max(vDistFxPtSolveNPLMC[vNumbFxPtSolveNPLMC>1]))
        colnames(mStatsDistFxPtSolveNPLMC) <- vStatsNames
        cat("Summary statistics of furthest equilibria:\n")
        print(as.matrix(mStatsDistFxPtSolveNPLMC))
      }
    }  
  } 
}


if (lSink==TRUE) sink()
if (lParallel==TRUE) stopCluster(dCl)

system("say Fini")  